'''
Contains Nonlinear Filename Wizard Class: autogenerates Filenames
'''

from PyQt5 import QtWidgets, QtGui, QtCore

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtProperty

from ..mesoSPIM_State import mesoSPIM_StateSingleton

import logging
logger = logging.getLogger(__name__)


class FilenameWizard(QtWidgets.QWizard):
    wizard_done = QtCore.pyqtSignal()

    num_of_pages = 6
    (welcome, raw, tiff, bigtiff, single_hdf5, finished) = range(num_of_pages)

    def __init__(self, parent=None):
        '''Parent is object of class mesoSPIM_AcquisitionManagerWindow()'''
        super().__init__(parent)

        ''' By an instance variable, callbacks to window signals can be handed
        through '''
        self.parent = parent
        self.state = mesoSPIM_StateSingleton()
        self.file_format = None  # 'raw', 'h5', 'tiff', 'btf'
        self.setWindowTitle('Filename Wizard')
        self.setPage(0, FilenameWizardWelcomePage(self))
        self.setPage(1, FilenameWizardRawSelectionPage(self))
        self.setPage(2, FilenameWizardTiffSelectionPage(self))
        self.setPage(3, FilenameWizardBigTiffSelectionPage(self))
        self.setPage(4, FilenameWizardSingleHDF5SelectionPage(self))
        self.setPage(5, FilenameWizardCheckResultsPage(self))
        self.setStyleSheet(''' font-size: 16px; ''')
        self.show()

    def done(self, r):
        ''' Reimplementation of the done function

        if r == 0: canceled
        if r == 1: finished properly
        '''
        if r == 0:
            logger.info('Filename Wizard was canceled')
        if r == 1:
            logger.info('Filename Wizard was closed properly')
            self.update_filenames_in_model()
        else:
            logger.info('Filename Wizard provided return code: ', r)

        super().done(r)

    def replace_spaces_with_underscores(self, string):
        return string.replace(' ','_')

    def replace_dots_with_underscores(self, string):
        return string.replace('.','_')

    def generate_filename_list(self, increment_number=True):
        '''
        Go through the model, entry for entry and populate the filenames
        '''
        row_count = self.parent.model.rowCount()
        num_string = '000000'
        start_number = 0
        start_number_string = str(start_number)
        self.filename_list = []
        for row in range(0, row_count):
            filename = ''
            if self.file_format == 'raw':
                if self.field('DescriptionRaw'):
                    filename += self.replace_spaces_with_underscores(self.field('DescriptionRaw')) + '_'

                if self.field('xyPosition'):
                    '''Round to nearest integer '''
                    x_position_string = str(int(round(self.parent.model.getXPosition(row))))
                    y_position_string = str(int(round(self.parent.model.getYPosition(row))))
                    filename += 'X' + x_position_string + '_' + 'Y' + y_position_string + '_'

                if self.field('rotationPosition'):
                    rot_position_string = str(int(round(self.parent.model.getRotationPosition(row))))
                    filename += 'rot_' + rot_position_string + '_'

                if self.field('Laser'):
                    filename += self.replace_spaces_with_underscores(self.parent.model.getLaser(row)) + '_'

                if self.field('Filter'):
                    filename += self.replace_spaces_with_underscores(self.parent.model.getFilter(row)) + '_'

                if self.field('Zoom'):
                    filename += self.replace_dots_with_underscores(self.parent.model.getZoom(row)) + '_'

                if self.field('Shutterconfig'):
                    filename += self.parent.model.getShutterconfig(row) + '_'

                file_suffix = num_string[:-len(start_number_string)] + start_number_string + '.' + self.file_format

                if increment_number:
                    start_number += 1
                    start_number_string = str(start_number)

            elif self.file_format in {'tiff', 'btf'}:
                if self.field('DescriptionTIFF'):
                    filename += self.replace_spaces_with_underscores(self.field('DescriptionTIFF')) + '_'

                if self.field('DescriptionBigTIFF'):
                    filename += self.replace_spaces_with_underscores(self.field('DescriptionBigTIFF')) + '_'

                if self.parent.model.getNShutterConfigs() > 1:
                    shutter_id = 0 if self.parent.model.getShutterconfig(row) == 'Left' else 1
                else:
                    shutter_id = 0
                filename += f'Tile{self.parent.model.getTileIndex(row)}_Ch{self.parent.model.getLaser(row)[:-3]}_Sh{shutter_id}'

                file_suffix = '.' + self.file_format

            elif self.file_format == 'h5':
                if self.field('DescriptionHDF5'):
                    filename += self.replace_spaces_with_underscores(self.field('DescriptionHDF5')) + '_'
                file_suffix = 'bdv.' + self.file_format

            else:
                raise ValueError(f"file suffix invalid: {self.file_format}")

            filename += file_suffix
            self.filename_list.append(filename)
            
    def update_filenames_in_model(self):
        row_count = self.parent.model.rowCount()
        filename_column = self.parent.model.getFilenameColumn()
        for row in range(0, row_count):
            filename = self.filename_list[row]
            index = self.parent.model.createIndex(row, filename_column)
            self.parent.model.setData(index, filename)


class FilenameWizardWelcomePage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setStyleSheet(''' font-size: 16px; ''')
        self.setTitle("Autogenerate filenames")
        self.setSubTitle("How would you like to save your data?")

        self.raw_string = 'Individual Raw Files: ~.raw'
        self.tiff_string = 'ImageJ TIFF files: ~.tiff'
        self.bigtiff_string = 'BigTIFF files: ~.btf'
        self.single_hdf5_string = 'BigDataViewer HDF5 file: ~.h5'

        self.SaveAsComboBoxLabel = QtWidgets.QLabel('Save as:')
        self.SaveAsComboBox = QtWidgets.QComboBox()
        self.SaveAsComboBox.addItems([self.raw_string, self.tiff_string, self.bigtiff_string, self.single_hdf5_string])
        self.SaveAsComboBox.setCurrentIndex(3)

        self.registerField('SaveAs', self.SaveAsComboBox, 'currentIndex')
        
        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.SaveAsComboBoxLabel, 0, 0)
        self.layout.addWidget(self.SaveAsComboBox, 0, 1)
        self.setLayout(self.layout)
    
    def nextId(self):
        if self.SaveAsComboBox.currentText() == self.raw_string:
            self.parent.file_format = 'raw'
            return self.parent.raw
        elif self.SaveAsComboBox.currentText() == self.tiff_string:
            self.parent.file_format = 'tiff'
            return self.parent.tiff
        elif self.SaveAsComboBox.currentText() == self.bigtiff_string:
            self.parent.file_format = 'btf'
            return self.parent.bigtiff
        elif self.SaveAsComboBox.currentText() == self.single_hdf5_string:
            self.parent.file_format = 'h5'
            return self.parent.single_hdf5


class AbstractSelectionPage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent
        self.setStyleSheet(''' font-size: 16px; ''')
        self.DescriptionCheckBox = QtWidgets.QCheckBox('Description: ', self)
        self.DescriptionCheckBox.setChecked(True)
        self.DescriptionLineEdit = QtWidgets.QLineEdit(self)
        self.DescriptionCheckBox.toggled.connect(lambda boolean: self.DescriptionLineEdit.setEnabled(boolean))
        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.DescriptionCheckBox, 0, 0)
        self.layout.addWidget(self.DescriptionLineEdit, 0, 1)
        self.setLayout(self.layout)

    def validatePage(self):
        self.parent.generate_filename_list()
        return super().validatePage()

    def nextId(self):
        return self.parent.finished


class FilenameWizardTiffSelectionPage(AbstractSelectionPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent
        self.setTitle("Autogenerate TIFF filenames")
        self.setSubTitle("Names will be compatible with BigStitcher auto-loader format:\n {Description}_Tile{}_Ch{}_Sh{}.tiff")
        self.registerField('DescriptionTIFF', self.DescriptionLineEdit)


class FilenameWizardBigTiffSelectionPage(AbstractSelectionPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Autogenerate BigTIFF filenames")
        self.setSubTitle("Names will be in format:\n {Description}_Tile{}_Ch{}_Sh{}.btf")
        self.registerField('DescriptionBigTIFF', self.DescriptionLineEdit)


class FilenameWizardRawSelectionPage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setStyleSheet(''' font-size: 16px; ''')
        self.setTitle("Autogenerate RAW filenames")
        self.setSubTitle("Which properties would you like to use?")

        self.DescriptionCheckBox = QtWidgets.QCheckBox('Description: ', self)
        self.DescriptionCheckBox.setChecked(True)
        self.DescriptionLineEdit = QtWidgets.QLineEdit(self)
        self.DescriptionCheckBox.toggled.connect(lambda boolean: self.DescriptionLineEdit.setEnabled(boolean))

        self.xyPositionCheckBox = QtWidgets.QCheckBox('XY Position')
        self.RotationPositionCheckBox = QtWidgets.QCheckBox('Rotation angle')

        self.LaserCheckBox = QtWidgets.QCheckBox('Laser', self)
        self.LaserCheckBox.setChecked(True)
        self.FilterCheckBox = QtWidgets.QCheckBox('Filter', self)
        # self.FilterCheckBox.setChecked(True)
        self.ZoomCheckBox = QtWidgets.QCheckBox('Zoom', self)
        self.ZoomCheckBox.setChecked(True)
        self.ShutterCheckBox = QtWidgets.QCheckBox('Shutterconfig', self)
        self.ShutterCheckBox.setChecked(True)

        self.registerField('DescriptionRaw', self.DescriptionLineEdit)
        self.registerField('xyPosition', self.xyPositionCheckBox)
        self.registerField('rotationPosition', self.RotationPositionCheckBox)
        self.registerField('Laser', self.LaserCheckBox)
        self.registerField('Filter', self.FilterCheckBox)
        self.registerField('Zoom', self.ZoomCheckBox)
        self.registerField('Shutterconfig', self.ShutterCheckBox)
        
        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.DescriptionCheckBox, 0, 0)
        self.layout.addWidget(self.DescriptionLineEdit, 0, 1)
        self.layout.addWidget(self.xyPositionCheckBox, 1, 0)
        self.layout.addWidget(self.RotationPositionCheckBox, 2, 0)
        self.layout.addWidget(self.LaserCheckBox, 3, 0)
        self.layout.addWidget(self.FilterCheckBox, 4, 0)
        self.layout.addWidget(self.ZoomCheckBox, 5, 0)
        self.layout.addWidget(self.ShutterCheckBox, 6, 0)
        self.setLayout(self.layout)

    def validatePage(self):
        self.parent.generate_filename_list()
        return super().validatePage()

    def nextId(self):
        return self.parent.finished


class FilenameWizardSingleHDF5SelectionPage(AbstractSelectionPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Autogenerate hdf5 filename")
        self.setSubTitle("This puts all raw data into a single hdf5 file, accompanied by two metadata files.")
        self.registerField('DescriptionHDF5', self.DescriptionLineEdit)

    def validatePage(self):
        self.parent.generate_filename_list(increment_number=False)
        return super().validatePage()


class FilenameWizardCheckResultsPage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setTitle('Check results')
        self.setSubTitle('Please check if the following filenames are ok:')

        self.TextEdit = QtWidgets.QPlainTextEdit(self)
        self.TextEdit.setReadOnly(True)

        self.mystring = ''        
        self.TextEdit.setPlainText(self.mystring)

        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.TextEdit, 0, 0, 1, 1)
        self.setLayout(self.layout)

    def initializePage(self):
        if self.parent.file_format in ('raw', 'tiff', 'btf'):
            file_list = self.parent.filename_list
        elif self.parent.file_format == 'h5':
            file_list = [self.parent.filename_list[0]]
        else:
            raise ValueError(f"file_format must be in ('raw', 'tiff', 'btf', 'h5'), received {self.parent.file_format}")

        for f in file_list:
            self.mystring += f
            self.mystring += '\n'
        self.TextEdit.setPlainText(self.mystring)        

    def cleanupPage(self):
        self.mystring = ''




for filter in ['515LP','561LP','594LP']:
	self.set_filter(filter, wait_until_done=True)
	time.sleep(0.1)
	
for zoom in ['1x','2x','4x','5x']:
	self.set_zoom(zoom, wait_until_done=True)
	time.sleep(0.1)

# Deprecated?
# for offset in np.linspace(0,5,21):
# 	self.set_galvo_l_offset(offset)
# 	time.sleep(0.2)


"""
mesoSPIM Module for controlling a shutter via NI-DAQmx
Author: Fabian Voigt
#TODO
"""

import nidaqmx
from nidaqmx.constants import LineGrouping

class NI_Shutter:
    """
    Slow shutter, intended more as a gating device than a fast open/close because the
    NI task is recreated and deleted every time a shutter is actuated.

    Thus, the shutter has more a "gating" function to protect the sample than
    fast control of the laser, this is done via the laser intensity anyway.

    This uses the property of NI-DAQmx-outputs to keep their last digital state or
    analog voltage for as long the device is not powered down.
    """
    def __init__(self, shutterline):
        self.shutterline = shutterline

        # Make sure that the Shutter is closed upon initialization
        with nidaqmx.Task() as task:
            task.do_channels.add_do_chan(self.shutterline, line_grouping=LineGrouping.CHAN_PER_LINE)
            task.write([False], auto_start=True)
            self.shutterstate = False

    # Open and close shutter take an optional argument to deal with the on_click method of Jupyter Widgets
    def open(self, *args):
        with nidaqmx.Task() as task:
            task.do_channels.add_do_chan(self.shutterline, line_grouping=LineGrouping.CHAN_PER_LINE)
            task.write([True], auto_start=True)
            self.shutterstate = True

    def close(self, *args):
        with nidaqmx.Task() as task:
            task.do_channels.add_do_chan(self.shutterline, line_grouping=LineGrouping.CHAN_PER_LINE)
            task.write([False], auto_start=True)
            self.shutterstate = False

    def state(self, *args):
        """ Returns "True" if the shutter is open, otherwise "False" """
        return self.shutterstate



'''
Serial thread for the mesoSPIM project
======================================

This thread handles all connections with serial devices such as stages,
filter wheels, zoom systems etc.
'''

import numpy as np
import time

import logging
logger = logging.getLogger(__name__)

'''PyQt5 Imports'''
from PyQt5 import QtWidgets, QtCore, QtGui

''' Import mesoSPIM modules '''
from .mesoSPIM_State import mesoSPIM_StateSingleton

#from .devices.filter_wheels.ludlcontrol import LudlFilterwheel
#from .devices.filter_wheels.mesoSPIM_FilterWheel import mesoSPIM_DemoFilterWheel
#from .devices.zoom.mesoSPIM_Zoom import DynamixelZoom, DemoZoom
#from .mesoSPIM_Stages import mesoSPIM_PIstage, mesoSPIM_DemoStage, mesoSPIM_GalilStages
# from .mesoSPIM_State import mesoSPIM_State

class mesoSPIM_Serial(QtCore.QObject):
    '''This class handles mesoSPIM serial connections'''
    sig_finished = QtCore.pyqtSignal()

    sig_state_request = QtCore.pyqtSignal(dict)
    
    sig_position = QtCore.pyqtSignal(dict)

    sig_zero_axes = QtCore.pyqtSignal(list)
    sig_unzero_axes = QtCore.pyqtSignal(list)
    sig_stop_movement = QtCore.pyqtSignal()
    sig_load_sample = QtCore.pyqtSignal()
    sig_unload_sample = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__()

        ''' Assign the parent class to a instance variable for callbacks '''
        self.parent = parent
        self.cfg = parent.cfg

        self.state = mesoSPIM_StateSingleton()

        ''' Handling of state changing requests '''
        self.parent.sig_state_request.connect(self.report_thread_id)
        self.parent.sig_state_request.connect(lambda dict: self.state_request_handler(dict, wait_until_done=False))

        # self.parent.sig_state_request_and_wait_until_done.connect(lambda dict: self.state_request_handler(dict, wait_until_done=True), type=3)

        # ''' Attaching the filterwheel '''
        # if self.cfg.filterwheel_parameters['filterwheel_type'] == 'Ludl':
        #     self.filterwheel = LudlFilterwheel(self.cfg.filterwheel_parameters['COMport'],self.cfg.filterdict)
        # elif self.cfg.filterwheel_parameters['filterwheel_type'] == 'DemoFilterWheel':
        #     self.filterwheel = mesoSPIM_DemoFilterWheel(self.cfg.filterdict)

        # ''' Attaching the zoom '''
        # if self.cfg.zoom_parameters['zoom_type'] == 'Dynamixel':
        #     self.zoom = DynamixelZoom(self.cfg.zoomdict,self.cfg.zoom_parameters['COMport'],self.cfg.zoom_parameters['servo_id'])
        # elif self.cfg.zoom_parameters['zoom_type'] == 'DemoZoom':
        #     self.zoom = DemoZoom(self.cfg.zoomdict)

        # ''' Attaching the stage '''
        # if self.cfg.stage_parameters['stage_type'] == 'PI':
        #     self.stage = mesoSPIM_PIstage(self)
        #     self.stage.sig_position.connect(lambda dict: self.sig_position.emit({'position': dict}))
        # elif self.cfg.stage_parameters['stage_type'] == 'GalilStage':
        #     self.stage = mesoSPIM_GalilStages(self)
        #     self.stage.sig_position.connect(lambda dict: self.sig_position.emit({'position': dict}))
        # elif self.cfg.stage_parameters['stage_type'] == 'DemoStage':
        #     self.stage = mesoSPIM_DemoStage(self)
        #     self.stage.sig_position.connect(lambda dict: self.sig_position.emit({'position': dict}))

        ''' Wiring signals through to child objects '''
        # self.parent.sig_move_relative.connect(lambda dict: self.move_relative(dict))
        # self.parent.sig_move_relative_and_wait_until_done.connect(lambda dict: self.move_relative(dict, wait_until_done=True), type=3)

        # self.parent.sig_move_absolute.connect(lambda dict: self.move_absolute(dict))
        # self.parent.sig_move_absolute_and_wait_until_done.connect(lambda dict: self.move_absolute(dict, wait_until_done=True), type=3)

        # self.parent.sig_zero_axes.connect(lambda list: self.sig_zero_axes.emit(list))
        # self.parent.sig_unzero_axes.connect(lambda list: self.sig_unzero_axes.emit(list))
        # self.parent.sig_stop_movement.connect(lambda: self.sig_stop_movement.emit())
        # self.parent.sig_load_sample.connect(self.sig_load_sample.emit)
        # self.parent.sig_unload_sample.connect(self.sig_unload_sample.emit)

        # self.parent.sig_go_to_rotation_position_and_wait_until_done.connect(lambda: self.go_to_rotation_position(wait_until_done=True), type=3)

        logger.info('Thread ID at Startup: '+str(int(QtCore.QThread.currentThreadId())))

    @QtCore.pyqtSlot(dict)
    def report_thread_id(self):
        logger.info('Demo Serial Thread ID: '+str(int(QtCore.QThread.currentThreadId())))

    @QtCore.pyqtSlot(dict)
    def state_request_handler(self, dict, wait_until_done=False):
        logger.info('Demo Serial Thread ID from request handler: '+str(int(QtCore.QThread.currentThreadId())))
        for key, value in zip(dict.keys(),dict.values()):
            # print('Serial thread: state request: Key: ', key, ' Value: ', value)
            '''
            Here, the request handling is done with lots if 'ifs'
            '''
            # print('Key: ', key, ' Value: ', value)
            if key == 'filter':
                if wait_until_done:
                    self.set_filter(value, wait_until_done)
                else:
                    self.set_filter(value)
            if key == 'zoom':
                if wait_until_done:
                    self.set_zoom(value, wait_until_done)
                else:
                    self.set_zoom(value)
            # Log Thread ID during Live: just debugging code
            if key == 'state':
                if value == 'live':
                    logger.info('Thread ID during live: '+str(int(QtCore.QThread.currentThreadId())))

    @QtCore.pyqtSlot(dict)
    def move_relative(self, dict, wait_until_done=False):
        logger.info('Thread ID during relative movement: '+str(int(QtCore.QThread.currentThreadId())))

        # logger.info('Thread ID during move rel: '+str(int(QtCore.QThread.currentThreadId())))
        # if wait_until_done:
        #     self.stage.move_relative(dict, wait_until_done=True)
        # else:
        #     self.stage.move_relative(dict)

    @QtCore.pyqtSlot(dict)
    def move_absolute(self, dict, wait_until_done=False):
        pass
        # if wait_until_done:
        #     self.stage.move_absolute(dict, wait_until_done=True)
        # else:
        #     self.stage.move_absolute(dict)

    @QtCore.pyqtSlot()
    def go_to_rotation_position(self, wait_until_done=False):
        logger.info('Thread ID during going to rotation position: '+str(int(QtCore.QThread.currentThreadId())))
        
        # if wait_until_done:
        #     self.stage.go_to_rotation_position(wait_until_done=True)
        # else:
        #     self.stage.go_to_rotation_position()

    @QtCore.pyqtSlot(str)
    def set_filter(self, filter, wait_until_done=False):
        logger.info('Thread ID during set filter: '+str(int(QtCore.QThread.currentThreadId())))
        # if wait_until_done:
        #     self.filterwheel.set_filter(filter, wait_until_done=True)
        # else:
        #     self.filterwheel.set_filter(filter, wait_until_done=False)
        self.state['filter'] = filter

    @QtCore.pyqtSlot(str)
    def set_zoom(self, zoom, wait_until_done=False):
        logger.info('Thread ID during set zoom: '+str(int(QtCore.QThread.currentThreadId())))
        # if wait_until_done:
        #     self.zoom.set_zoom(zoom, wait_until_done=True)
        # else:
        #     self.zoom.set_zoom(zoom, wait_until_done=False)
        self.state['zoom'] = zoom
        self.state['pixelsize'] = self.cfg.pixelsize[zoom]


"""
mesoSPIM Module for controlling ASI-Stages 

Author: Fabian Voigt
"""

import time
import traceback
import serial

from PyQt5 import QtWidgets, QtCore, QtGui

import logging
logger = logging.getLogger(__name__)


class StageControlASITiger(QtCore.QObject):
    sig_pause = QtCore.pyqtSignal(bool)

    '''
    Class to control a ASI Tiger mechanical stage controller
    
    Inherits from QtCore.QObject so it can be moved to a QThread.

    Note:
        This is a custom ASI stageset which contains the axis designations:
        X Y Z T V W 
    '''

    def __init__(self, asi_parameters):
        super().__init__()
        self.port = asi_parameters['COMport']
        self.baudrate = asi_parameters['baudrate']
        self.stage_assignment = asi_parameters['stage_assignment']
        self.axis_list = [self.stage_assignment[i] for i in self.stage_assignment.keys()]
        self.axes = ''.join(self.axis_list) # String containing all axes
        self.num_axes = len(self.axes) # The number of axes
        self.encoder_conversion = asi_parameters['encoder_conversion']
        
        self.position_dict = {axis : None for axis in self.axis_list} # create an empty position dict
                
        '''Open connection to the stage controller'''
        self.asi_connection = serial.Serial(self.port, self.baudrate, parity=serial.PARITY_NONE, timeout=5, xonxoff=False, stopbits=serial.STOPBITS_ONE)
        self.previous_command = ''
        self.current_z_slice = 0

    def close(self):
        '''Closes connection to the stage'''
        self.asi_connection.close()
    
    def _send_command(self, command: bytes):
        '''Sends a command to the controller
        Try-except block included to catch errors - this is dangerous - no checking of success
        Args:
            command (bytes): Command string to be sent. Needs to be in the form "b'W V?\r'" - has to be binary, do not forget the carriage return
            
        Returns:
            answer (str): Answer by the controller.
        '''
        try:
            ''' During acquisitions: send pause signal '''
            self.sig_pause.emit(True)
            start_time = time.time()
            self._reset_buffers()
            logger.debug(f"Serial sent: {command}")
            self.asi_connection.write(command)
            message = self.asi_connection.readline().decode("ascii")
            response_time = time.time() 
            ''' During acquistions: send unpause signal '''
            self.sig_pause.emit(False)
            logger.debug(f"Serial received: {message} ")
            ''' Logging of serial connections if response >30 ms (previously: 15 ms)'''
            delta_t = round(response_time - start_time, 6)
            if delta_t > 0.04:
                logger.info('Z-Slice (only valid during acq): ' + str(self.current_z_slice) + ' Response time, s (if >0.04): ' + str(delta_t))
            return message
        except Exception as error:
            logger.error(f"Serial exception of the ASI stage: command {command.decode('ascii')}, error: {error}")

    def _reset_buffers(self):
        if self.asi_connection is not None:
            self.asi_connection.reset_input_buffer()
            self.asi_connection.reset_output_buffer()
        else:
            logger.error("Serial port not initialized")

    def axis_in_config_check(self, axis):
        '''
        Checks if a axis string is in self.axes
        
        Returns:
            True if axis in config
        '''
        if axis in self.axes:
            return True 
        else:
            print('Axis: ', axis, ' not found in ASI stage config.')
            logger.info('Axis: ' + str(axis) + ' not found in ASI stage config.')
            return False
                  
    def stop(self):
        '''Stops movement on all axes by sending "halt" to the controller
        '''
        response = self._send_command(b'\\r')
        logger.info(f"ASI response to HALT command: {response}")
        
    def wait_until_done(self):
        '''Blocks if the stage is moving due to a serial command'''

        '''If the stage returns 'B'as the first letter, it is busy, if it returns 'N', it is done.
        Only if the stage returns 'N' twice, it is not busy.
        '''
        self.stage_busy = True
        while self.stage_busy is True:
            try: 
                message1 = self._send_command(b'/\r')[0]
                time.sleep(0.05)
                message2 = self._send_command(b'/\r')[0]
                time.sleep(0.05)
                if message1 == 'N' and message2 == 'N':
                    self.stage_busy = False
            except:
                logger.error('ASI stages: Wait until done failed')
        
    def read_position(self):
        '''Reports position from the stages
        Returns:
            positions (dictionary): list of positions 
        
        '''
        command_string = 'W ' + self.axes + '\r'
        position_string = self._send_command(command_string.encode('ascii'))
        # Create a list of the form "['7835', '-38704', '0', '0', '-367586']", first element is the ack ':A' and gets discarded
        if position_string is not None:
            try:
                position_list = position_string.split()[1:]
                ''' Only process position list if it contains all values'''
                if len(position_list) == self.num_axes:
                    try:
                        position_list = [int(value) for value in position_list]
                        endcoder_conversion_list = list(self.encoder_conversion.values())
                        position_dict = {self.axes[i]: position_list[i]/endcoder_conversion_list[i] for i in range(self.num_axes)}
                        if position_dict is not None:
                            self.position_dict = position_dict
                            return position_dict
                    except:
                        logger.info('Invalid position dict: ' + str(position_list))
                        # return last position dict
                        return self.position_dict
                else:
                    logger.error(f"Position list count {position_list} does not match the number of axes {self.num_axes}")
            except: 
                logger.error('Invalid position string: ' + str(position_string))
        else:
            logger.error("Position string is empty")

    def move_relative(self, motion_dict):
        '''Command for relative motion 
        
        Args:
            motion_dict (dict): Dictionary in the form {1: 4000, 2:-234}, in general {axis_id:requested_position} 
                                with axis_id (str) and requested_position (int) in um.
        Returns:
            Carries out motion
        '''
        command_string = 'R'
        for axis in motion_dict.keys():
            if self.axis_in_config_check(axis):
                command_string = command_string + ' ' + axis + '=' + str(motion_dict[axis]*self.encoder_conversion[axis])
        if command_string != 'R':
            command_string += '\r'
            self._send_command(command_string.encode('ascii'))
        
    def move_absolute(self, motion_dict):
        ''' Command for absolute motion 
        
        Args:
            motion_dict (dict): Dictionary in the form {1: 4000, 2:-234}, in general {axis_id:requested_position} 
                                with axis_id (str) and requested_position (int) in um.
        
        '''
        command_string = 'M'
        for axis in motion_dict.keys():
            if self.axis_in_config_check(axis):
                command_string = command_string + ' ' + axis + '=' + str(motion_dict[axis]*self.encoder_conversion[axis])
                
        if command_string != 'M':
            command_string += '\r'
            self._send_command(command_string.encode('ascii'))

    def enable_ttl_mode(self, card_ids, bool):
        ''' Enables or disables TTL mode of ASI controllers
        Args:
            card_ids (list): List of card IDs inside the controller (i.e. (2,3) for cards in slots 2 and 3) for 
                            which TTL triggering should be enabled or disabled. If None, the controller is assumed
                            not to have any card slots, i.e. an MS-2000 controller
            
            bool (boolean): True or False depending on whether TTL mode should be enabled or disabled.
        '''
        if card_ids is not None: # Tiger controller
            if bool is True: # Enable TTL mode for all cards
                for i in card_ids:
                    command_string = str(i) + ' TTL X=2 Y=2\r'
                    self._send_command(command_string.encode('ascii'))
                    logger.info('TTL enabled')
            else: # Disable TTL mode for all cards
                for i in card_ids:
                    command_string = str(i) + ' TTL X=0 Y=2\r'
                    self._send_command(command_string.encode('ascii'))
                    logger.info('TTL disabled')
        else: # MS-2000 controller
            if bool is True:
                self._send_command(b'TTL X=2 Y=2\r') # MS-2000 TTL mode should be enabled
                logger.info('TTL enabled')
            else:
                self._send_command(b'TTL X=0 Y=2\r') # MS-2000 TTL mode should be disabled
                logger.info('TTL disabled')



for i in range(10):
	print('Starting movement')
	self.serial_worker.move_relative({'z_rel':1000}, wait_until_done=True)
	print('Done with the first movement')
	time.sleep(1)
	print('Starting second movement')
	self.serial_worker.move_relative({'z_rel':-1000}, wait_until_done=True)
	time.sleep(1)
	print('Done with the second movement')


'''
mesoSPIM MainWindow
'''
import tifffile
import logging
import time
logger = logging.getLogger(__name__)
logger.setLevel('INFO') # set this to INFO to avoid excessive GUI logging messages, DEBUG for printing thread stats.

from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.uic import loadUi

''' Disabled taskbar button progress display due to problems with Anaconda default'''
# if sys.platform == 'win32':
#     from PyQt5.QtWinExtras import QWinTaskbarButton

from .mesoSPIM_CameraWindow import mesoSPIM_CameraWindow
from .mesoSPIM_AcquisitionManagerWindow import mesoSPIM_AcquisitionManagerWindow
from .mesoSPIM_Serial import mesoSPIM_Serial
from .mesoSPIM_Optimizer import mesoSPIM_Optimizer
from .WebcamWindow import WebcamWindow
from .mesoSPIM_ContrastWindow import mesoSPIM_ContrastWindow
from .mesoSPIM_ScriptWindow import mesoSPIM_ScriptWindow # do not delete this line, it is actually used in exec()

from .mesoSPIM_State import mesoSPIM_StateSingleton
from .mesoSPIM_Core import mesoSPIM_Core
from .devices.joysticks.mesoSPIM_JoystickHandlers import mesoSPIM_JoystickHandler


class LogDisplayHandler(QtCore.QObject, logging.Handler):
    """ Handler class to display log in a TextDisplay widget. A thread-safe version, callable from non-GUI threads."""
    new_record = QtCore.pyqtSignal(object)

    def __init__(self, parent):
        super().__init__(parent)
        super(logging.Handler).__init__()
        formatter = Formatter('%(levelname)s:%(module)s:%(funcName)s:%(message)s')
        self.setFormatter(formatter)

    def emit(self, record):
        msg = self.format(record)
        self.new_record.emit(msg)  # <---- emit signal here


class Formatter(logging.Formatter):
    """ Formatter of the LogDisplayHandler class."""
    def formatException(self, ei):
        result = super(Formatter, self).formatException(ei)
        return result

    def format(self, record):
        s = super(Formatter, self).format(record)
        if record.exc_text:
            s = s.replace('\n', '')
        return s


class mesoSPIM_MainWindow(QtWidgets.QMainWindow):
    """ Main application window which instantiates worker objects and moves them to a thread. """
    # sig_live = QtCore.pyqtSignal()
    sig_stop = QtCore.pyqtSignal()
    sig_finished = QtCore.pyqtSignal()
    sig_enable_gui = QtCore.pyqtSignal(bool)
    sig_state_request = QtCore.pyqtSignal(dict)
    sig_execute_script = QtCore.pyqtSignal(str)
    sig_move_relative = QtCore.pyqtSignal(dict)
    # sig_move_relative_and_wait_until_done = QtCore.pyqtSignal(dict)
    sig_move_absolute = QtCore.pyqtSignal(dict)
    # sig_move_absolute_and_wait_until_done = QtCore.pyqtSignal(dict)
    sig_zero_axes = QtCore.pyqtSignal(list)
    sig_unzero_axes = QtCore.pyqtSignal(list)
    sig_stop_movement = QtCore.pyqtSignal()
    sig_load_sample = QtCore.pyqtSignal()
    sig_unload_sample = QtCore.pyqtSignal()

    sig_save_etl_config = QtCore.pyqtSignal()
    sig_poke_demo_thread = QtCore.pyqtSignal()
    sig_launch_optimizer = QtCore.pyqtSignal(dict)
    sig_launch_contrast_window = QtCore.pyqtSignal()

    def __init__(self, package_directory, config, title="mesoSPIM Main Window"):
        super().__init__()
        # Initial housekeeping
        self.cfg = config
        self.package_directory = package_directory
        self.script_window_counter = 0
        self.update_gui_from_state_flag = False

        # Instantiate the one and only mesoSPIM state '''
        self.state = mesoSPIM_StateSingleton()
        self.state.sig_updated.connect(self.update_gui_from_state)

        # Setting up the user interface windows
        loadUi(self.package_directory + '/gui/mesoSPIM_MainWindow.ui', self)
        self.setWindowTitle(title)

        # Connect log display widget
        self.log_display_handler = LogDisplayHandler(self)
        self.log_display_handler.new_record.connect(self.LogTextDisplay.appendPlainText)

        self.camera_window = mesoSPIM_CameraWindow(self)
        self.camera_window.show()

        self.acquisition_manager_window = mesoSPIM_AcquisitionManagerWindow(self)
        self.acquisition_manager_window.show()

        self.webcam_window = None

        # arrange the windows on the screen, tiled
        if hasattr(self.cfg, 'ui_options') and 'window_pos' in self.cfg.ui_options.keys():
            window_pos = self.cfg.ui_options['window_pos']
        else:
            window_pos = (100, 100)
        self.move(window_pos[0], window_pos[1])
        self.camera_window.move(window_pos[0] + self.width() + 50, window_pos[1])
        self.acquisition_manager_window.move(window_pos[0], window_pos[1] + self.height() + 50)
        if self.webcam_window:
            self.webcam_window.move(window_pos[0] + self.width() + self.camera_window.width() + 50, window_pos[1])

        # set up some Acq manager signals
        self.acquisition_manager_window.sig_warning.connect(self.display_warning)
        self.acquisition_manager_window.sig_move_absolute.connect(self.sig_move_absolute.emit)

        # Setting up the threads
        logger.info('Ideal thread count: '+str(int(QtCore.QThread.idealThreadCount())))
        self.core_thread = QtCore.QThread()
        # Entry point: Work on thread affinity here
        self.core = mesoSPIM_Core(self.cfg, self)

        self.core.moveToThread(self.core_thread)
        self.core.waveformer.moveToThread(self.core_thread)
        # This shit below does not work, everything remains in the main thread. Why?
        # Even if "moved" to core thread, all these objects actually run in the main thread.
        # It is a working solution, but not ideal. Ideally these workers must run in self.core_thread.
        #self.core.serial_worker.moveToThread(self.core_thread)
        #self.core.serial_worker.stage.moveToThread(self.core_thread)
        #self.core.serial_worker.stage.asi_stages.moveToThread(self.core_thread)
        #self.core.serial_worker.stage.pos_timer.moveToThread(self.core_thread)

        # Get buttons & connections ready
        self.initialize_and_connect_menubar()
        self.initialize_and_connect_widgets()

        # Widget list for blockSignals during status updates
        self.widgets_to_block = []
        self.parent_widgets_to_block = [self.ETLTabWidget, self.ParameterTabWidget, self.ControlGroupBox]
        self.create_widget_list(self.parent_widgets_to_block, self.widgets_to_block)

        # The signal switchboard, Core -> MainWindow
        self.core.sig_finished.connect(self.finished)
        self.core.sig_position.connect(self.update_position_indicators)
        self.core.sig_update_gui_from_state.connect(self.enable_gui_updates_from_state)
        self.core.sig_status_message.connect(self.display_status_message)
        self.core.sig_progress.connect(self.update_progressbars)
        self.core.sig_warning.connect(self.display_warning)

        self.optimizer = None
        self.contrast_window = None

        # The signal switchboard, MainWindow -> Core
        self.sig_launch_optimizer.connect(self.launch_optimizer)
        self.sig_launch_contrast_window.connect(self.launch_contrast_window)

        ''' Start the thread '''
        self.core_thread.start(QtCore.QThread.HighPriority)

        try:
            self.thread().setPriority(QtCore.QThread.HighestPriority)
            logger.debug('Main Window Thread priority: '+str(self.thread().priority()))
        except:
            logger.error(f'Main Window: Printing Thread priority failed.')

        logger.debug(f'Main Window: Core priority: {self.core_thread.priority()}')

        self.joystick = mesoSPIM_JoystickHandler(self)

        self.enable_gui_updates_from_state(False)

    def open_webcam_window(self):
        """Open USB webcam window using cam ID specified in config file. Otherwise, try to open with ID=0"""
        if self.webcam_window is None: # first call
            if hasattr(self.cfg, 'ui_options') and ('usb_webcam_ID' in self.cfg.ui_options.keys()):
                if self.cfg.ui_options['usb_webcam_ID'] >= 0:
                    self.webcam_window = WebcamWindow(self.cfg.ui_options['usb_webcam_ID'])
                    self.webcam_window.show()
            else:
                self.webcam_window = WebcamWindow(0)
                self.webcam_window.show()
        else: # open previously close window
            self.webcam_window.show()

    def __del__(self):
        """Cleans the threads up after deletion, waits until the threads
        have truly finished their life.

        Make sure to keep this up to date with the number of threads
        """
        try:
            self.core_thread.quit()
            self.core_thread.wait()
        except:
            pass

    def close_app(self):
        self.camera_window.close()
        self.acquisition_manager_window.close()
        if self.optimizer:
            self.optimizer.close()
        try:
            self.webcam_window.close()
        except:
            pass
        if self.contrast_window:
            self.contrast_window.close()
        self.close()

    def open_tiff(self):
        """Open and display a TIFF file (stack), eg for demo and debugging purposes."""
        tiff_path, _ = QtWidgets.QFileDialog.getOpenFileName(None, 'Open TIFF', "./", "TIFF files (*tif; *tiff)")
        if tiff_path:
            try:
                stack = tifffile.imread(tiff_path)
                self.camera_window.set_image(stack)
                logger.info(f"Loaded TIFF file from {tiff_path}, dimensions {stack.shape}")
            except Exception as e:
                logger.exception(f"{e}")
        else:
            logger.info(f"Loaded TIFF file path is None")

    def get_state_parameter(self, state_parameter):
        return self.state[state_parameter]

    def check_instances(self, widget):
        ''' 
        Method to check whether a widget belongs to one of the Qt Classes specified.

        Args:
            widget (QtWidgets.QWidget): Widget to check 

        Returns:
            return_value (bool): True if widget is in the list, False if not.
        '''
        if isinstance(widget, (QtWidgets.QSpinBox, 
                                QtWidgets.QDoubleSpinBox,
                                QtWidgets.QSlider,
                                QtWidgets.QComboBox,
                                QtWidgets.QPushButton)):
            return True 
        else:
            return False

    def create_widget_list(self, list, widget_list):
        """
        Helper method to recursively loop through all the widgets in a list and their children.
        Args:
            list (list): List of QtWidgets.QWidget objects
        """
        for widget in list:
            if list != ([] or None):
                if self.check_instances(widget):
                    # print(widget.objectName())
                    widget_list.append(widget)
                list = widget.children()
                self.create_widget_list(list, widget_list)
            else:
                return None

    @QtCore.pyqtSlot(str)
    def display_status_message(self, string):
        """
        Displays a message in the status bar for a time in ms
        """
        self.statusBar().showMessage(string)

    def pos2str(self, position):
        """ Little helper method for converting positions to strings """
        return '%.1f' % position

    @QtCore.pyqtSlot(dict)
    def update_position_indicators(self, dict):
        for key, pos_dict in dict.items():
            if key == 'position':
                self.x_position, self.y_position, self.z_position = pos_dict['x_pos'], pos_dict['y_pos'], pos_dict['z_pos']
                self.f_position, self.theta_position = pos_dict['f_pos'], pos_dict['theta_pos']
                self.X_Position_Indicator.setText(self.pos2str(self.x_position)+' µm')
                self.Y_Position_Indicator.setText(self.pos2str(self.y_position)+' µm')
                self.Z_Position_Indicator.setText(self.pos2str(self.z_position)+' µm')
                self.Focus_Position_Indicator.setText(self.pos2str(self.f_position)+' µm')
                self.Rotation_Position_Indicator.setText(self.pos2str(self.theta_position)+'°')
                self.state['position'] = dict['position']

    @QtCore.pyqtSlot(dict)
    def update_progressbars(self,dict):
        cur_acq = dict['current_acq']
        tot_acqs = dict['total_acqs']
        cur_image = dict['current_image_in_acq']
        images_in_acq = dict['images_in_acq']
        tot_images = dict['total_image_count']
        image_count = dict['image_counter']
        time_passed_string = dict['time_passed_string']
        remaining_time_string = dict['remaining_time_string']

        self.AcquisitionProgressBar.setValue(int((cur_image+1)/images_in_acq*100))
        self.TotalProgressBar.setValue(int((image_count+1)/tot_images*100))

        ''' Disabled taskbar button progress display due to problems with Anaconda default
        if sys.platform == 'win32':
            self.win_taskbar_button.progress().setValue(int((image_count+1)/tot_images*100))
        '''

        self.AcquisitionProgressBar.setFormat('%p% Image '+ str(cur_image+1) +\
                                        '/' + str(images_in_acq) + ' ')
        self.TotalProgressBar.setFormat('%p% Acq: '+ str(cur_acq+1) +\
                                        '/' + str(tot_acqs) +\
                                         ' ' + ' Image: '+ str(image_count) +\
                                        '/' + str(tot_images) + ' ' +\
                                            'Time: ' + time_passed_string + \
                                            ' Remaining: ' + remaining_time_string)

    def create_script_window(self):
        """
        Creates a script window and binds it to a self.scriptwindow0 ... n instanceself.

        This happens dynamically using exec which should be replaced at
        some point with a factory pattern.
        """
        windowstring = 'self.scriptwindow'+str(self.script_window_counter)
        exec(windowstring+ '= mesoSPIM_ScriptWindow(self)')
        exec(windowstring+'.setWindowTitle("Script Window #'+str(self.script_window_counter)+'")')
        exec(windowstring+'.show()')
        exec(windowstring+'.sig_execute_script.connect(self.execute_script)')
        self.script_window_counter += 1

    def initialize_and_connect_menubar(self):
        self.actionExit.triggered.connect(self.close_app)
        self.actionOpen_TIFF.triggered.connect(self.open_tiff)
        self.actionOpen_Camera_Window.triggered.connect(self.camera_window.show)
        self.actionOpen_Webcam_Window.triggered.connect(self.open_webcam_window)
        self.actionOpen_Acquisition_Manager.triggered.connect(self.acquisition_manager_window.show)
        self.actionCascade_windows.triggered.connect(self.cascade_all_windows)

    def initialize_and_connect_widgets(self):
        """ Connecting the menu actions """
        self.openScriptEditorButton.clicked.connect(self.create_script_window)

        ''' Connecting the movement & zero buttons '''
        self.xPlusButton.pressed.connect(lambda: self.move_relative({'x_rel': -self.xyzIncrementSpinbox.value()}))
        self.xMinusButton.pressed.connect(lambda: self.move_relative({'x_rel': self.xyzIncrementSpinbox.value()}))
        self.yPlusButton.pressed.connect(lambda: self.move_relative({'y_rel': self.xyzIncrementSpinbox.value()}))
        self.yMinusButton.pressed.connect(lambda: self.move_relative({'y_rel': -self.xyzIncrementSpinbox.value()}))
        self.zPlusButton.pressed.connect(lambda: self.move_relative({'z_rel': self.xyzIncrementSpinbox.value()}))
        self.zMinusButton.pressed.connect(lambda: self.move_relative({'z_rel': -self.xyzIncrementSpinbox.value()}))
        self.focusPlusButton.pressed.connect(lambda: self.move_relative({'f_rel': self.focusIncrementSpinbox.value()}))
        self.focusMinusButton.pressed.connect(lambda: self.move_relative({'f_rel': -self.focusIncrementSpinbox.value()}))
        self.rotPlusButton.pressed.connect(lambda: self.move_relative({'theta_rel': self.rotIncrementSpinbox.value()}))
        self.rotMinusButton.pressed.connect(lambda: self.move_relative({'theta_rel': -self.rotIncrementSpinbox.value()}))

        self.xyzrotStopButton.pressed.connect(self.sig_stop_movement.emit)

        self.xyZeroButton.clicked.connect(lambda bool: self.sig_zero_axes.emit(['x','y']) if bool is True else self.sig_unzero_axes.emit(['x','y']))
        self.zZeroButton.clicked.connect(lambda bool: self.sig_zero_axes.emit(['z']) if bool is True else self.sig_unzero_axes.emit(['z']))
        self.focusZeroButton.clicked.connect(lambda bool: self.sig_zero_axes.emit(['f']) if bool is True else self.sig_unzero_axes.emit(['f']))
        self.focusAutoButton.clicked.connect(lambda: self.sig_launch_optimizer.emit({'mode': 'focus', 'amplitude': 300}))
        self.rotZeroButton.clicked.connect(lambda bool: self.sig_zero_axes.emit(['theta']) if bool is True else self.sig_unzero_axes.emit(['theta']))
        self.xyzLoadButton.clicked.connect(self.sig_load_sample.emit)
        self.xyzUnloadButton.clicked.connect(self.sig_unload_sample.emit)
        self.launchOptimizerButton.clicked.connect(lambda: self.sig_launch_optimizer.emit({'mode': 'etl_offset', 'amplitude': 0.5}))
        self.ContrastWindowButton.clicked.connect(lambda: self.sig_launch_contrast_window.emit())

        ''' Disabling UI buttons if necessary '''
        if hasattr(self.cfg, 'ui_options'):
            if self.cfg.ui_options['enable_x_buttons'] is False:
                self.xPlusButton.setEnabled(False)
                self.xMinusButton.setEnabled(False)

            if self.cfg.ui_options['enable_y_buttons'] is False:
                self.yPlusButton.setEnabled(False)
                self.yMinusButton.setEnabled(False)

            if self.cfg.ui_options['enable_x_buttons'] is False and self.cfg.ui_options['enable_y_buttons'] is False:
                self.xyZeroButton.setEnabled(False)

            if self.cfg.ui_options['enable_z_buttons'] is False:
                self.zPlusButton.setEnabled(False)
                self.zMinusButton.setEnabled(False)
                self.zZeroButton.setEnabled(False)

            if self.cfg.ui_options['enable_f_buttons'] is False:
                self.focusPlusButton.setEnabled(False)
                self.focusMinusButton.setEnabled(False)
                self.focusZeroButton.setEnabled(False)

            if self.cfg.ui_options['enable_rotation_buttons'] is False:
                self.rotPlusButton.setEnabled(False)
                self.rotMinusButton.setEnabled(False)
                self.rotZeroButton.setEnabled(False)

            if self.cfg.ui_options['enable_loading_buttons'] is False:
                self.xyzLoadButton.setEnabled(False)
                self.xyzUnloadButton.setEnabled(False)

        ''' Connecting state-changing buttons '''
        self.LiveButton.clicked.connect(self.run_live)
        self.SnapButton.clicked.connect(self.run_snap)
        self.RunSelectedAcquisitionButton.clicked.connect(self.run_selected_acquisition)
        self.RunAcquisitionListButton.clicked.connect(self.run_acquisition_list)
        self.StopButton.clicked.connect(lambda: self.sig_state_request.emit({'state':'idle'}))
        #self.StopButton.clicked.connect(lambda: print('Stopping'))
        self.LightsheetSwitchingModeButton.clicked.connect(self.run_lightsheet_alignment_mode)
        self.VisualModeButton.clicked.connect(self.run_visual_mode)

        self.ETLIncrementSpinBox.valueChanged.connect(self.update_etl_increments)
        self.ZeroLeftETLButton.toggled.connect(self.zero_left_etl)
        self.ZeroRightETLButton.toggled.connect(self.zero_right_etl)
        self.freezeGalvoButton.toggled.connect(self.zero_galvo_amp)

        self.ChooseETLcfgButton.clicked.connect(self.choose_etl_config)
        self.SaveETLParametersButton.clicked.connect(self.save_etl_config)

        self.ChooseSnapFolderButton.clicked.connect(self.choose_snap_folder)
        self.SnapFolderIndicator.setText(self.state['snap_folder'])
        self.ETLconfigIndicator.setText(self.state['ETL_cfg_file'])

        self.widget_to_state_parameter_assignment=(
            (self.FilterComboBox, 'filter',1),
            (self.FilterComboBox, 'filter',1),
            (self.ZoomComboBox, 'zoom',1),
            (self.ShutterComboBox, 'shutterconfig',1),
            (self.LaserComboBox, 'laser',1),
            (self.LaserIntensitySlider, 'intensity', 1),
            (self.LaserIntensitySpinBox, 'intensity', 1),
            (self.CameraExposureTimeSpinBox, 'camera_exposure_time',1000),
            #(self.CameraLineIntervalSpinBox,'camera_line_interval',1000000),
            (self.CameraTriggerDelaySpinBox,'camera_delay_%',1),
            (self.CameraTriggerPulseLengthSpinBox, 'camera_pulse_%',1),
            (self.SweeptimeSpinBox,'sweeptime',1000),
            (self.LeftLaserPulseDelaySpinBox,'laser_l_delay_%',1),
            (self.RightLaserPulseDelaySpinBox,'laser_r_delay_%',1),
            (self.LeftLaserPulseLengthSpinBox,'laser_l_pulse_%',1),
            (self.RightLaserPulseLengthSpinBox,'laser_r_pulse_%',1),
            (self.LeftLaserPulseMaxAmplitudeSpinBox,'laser_l_max_amplitude_%',1),
            (self.RightLaserPulseMaxAmplitudeSpinBox,'laser_r_max_amplitude_%',1),
            (self.GalvoFrequencySpinBox,'galvo_r_frequency',1),
            (self.GalvoFrequencySpinBox,'galvo_l_frequency',1),
            (self.LeftGalvoAmplitudeSpinBox,'galvo_l_amplitude',1),
            (self.LeftGalvoAmplitudeSpinBox,'galvo_r_amplitude',1),
            (self.LeftGalvoPhaseSpinBox,'galvo_l_phase',1),
            (self.RightGalvoPhaseSpinBox,'galvo_r_phase',1),
            (self.LeftGalvoOffsetSpinBox, 'galvo_l_offset',1),
            (self.RightGalvoOffsetSpinBox, 'galvo_r_offset',1),
            (self.LeftETLOffsetSpinBox,'etl_l_offset',1),
            (self.RightETLOffsetSpinBox,'etl_r_offset',1),
            (self.LeftETLAmplitudeSpinBox,'etl_l_amplitude',1),
            (self.RightETLAmplitudeSpinBox,'etl_r_amplitude',1),
            (self.LeftETLDelaySpinBox,'etl_l_delay_%',1),
            (self.RightETLDelaySpinBox,'etl_r_delay_%',1),
            (self.LeftETLRampRisingSpinBox,'etl_l_ramp_rising_%',1),
            (self.RightETLRampRisingSpinBox, 'etl_r_ramp_rising_%',1),
            (self.LeftETLRampFallingSpinBox, 'etl_l_ramp_falling_%',1),
            (self.RightETLRampFallingSpinBox, 'etl_r_ramp_falling_%',1)
        )

        for widget, state_parameter, conversion_factor in self.widget_to_state_parameter_assignment:
            self.connect_widget_to_state_parameter(widget, state_parameter, conversion_factor)

        ''' Connecting the microscope controls '''

        ''' List for subsampling factors - comboboxes need a list of strings'''
        subsampling_list = [str(i) for i in self.cfg.camera_parameters['subsampling']]

        self.connect_combobox_to_state_parameter(self.FilterComboBox,self.cfg.filterdict.keys(),'filter')
        self.connect_combobox_to_state_parameter(self.ZoomComboBox,self.cfg.zoomdict.keys(),'zoom')
        self.connect_combobox_to_state_parameter(self.ShutterComboBox,self.cfg.shutteroptions,'shutterconfig')
        self.connect_combobox_to_state_parameter(self.LaserComboBox,self.cfg.laserdict.keys(),'laser')
        # self.connect_combobox_to_state_parameter(self.CameraSensorModeComboBox,['ASLM','Area'],'camera_sensor_mode')
        self.connect_combobox_to_state_parameter(self.LiveSubSamplingComboBox,subsampling_list,'camera_display_live_subsampling', int_conversion = True)
        self.connect_combobox_to_state_parameter(self.AcquisitionSubSamplingComboBox,subsampling_list,'camera_display_acquisition_subsampling', int_conversion = True)
        # self.connect_combobox_to_state_parameter(self.CameraSensorModeComboBox,['ASLM','Area'],'camera_sensor_mode')
        self.connect_combobox_to_state_parameter(self.BinningComboBox, self.cfg.binning_dict.keys(),'camera_binning')

        self.LaserIntensitySlider.valueChanged.connect(self.set_laser_intensity)
        self.LaserIntensitySpinBox.valueChanged.connect(self.set_laser_intensity)
        self.LaserIntensitySlider.setValue(self.cfg.startup['intensity'])

    @QtCore.pyqtSlot(dict)
    def move_relative(self, pos_dict):
        assert len(pos_dict) == 1, f"Position dictionary expects only one entry, got {pos_dict}"
        key, value = list(pos_dict.keys())[0], list(pos_dict.values())[0]
        if key == 'x_rel':
            if not (self.cfg.stage_parameters['x_min'] + 1000 <= self.x_position + value <= self.cfg.stage_parameters['x_max'] - 1000):
                self.X_Position_Indicator.setStyleSheet("color: red;")
                logger.info(f"X axis position {self.x_position} close to software limits, defined in the config file.")
            else:
                self.X_Position_Indicator.setStyleSheet("color: white;")
        elif key == 'y_rel':
            if not (self.cfg.stage_parameters['y_min'] + 1000 <= self.y_position + value <= self.cfg.stage_parameters['y_max'] - 1000):
                self.Y_Position_Indicator.setStyleSheet("color: red;")
                logger.info(f"Y axis position {self.y_position} close to software limits, defined in the config file.")
            else:
                self.Y_Position_Indicator.setStyleSheet("color: white;")
        elif key == 'z_rel':
            if not (self.cfg.stage_parameters['z_min'] + 1000 <= self.z_position + value <= self.cfg.stage_parameters['z_max'] - 1000):
                self.Z_Position_Indicator.setStyleSheet("color: red;")
                logger.info(f"Z axis position {self.z_position} close to software limits, defined in the config file.")
            else:
                self.Z_Position_Indicator.setStyleSheet("color: white;")
        elif key == 'f_rel':
            if not (self.cfg.stage_parameters['f_min'] + 1000 <= self.f_position + value <= self.cfg.stage_parameters['f_max'] - 1000):
                logger.info(f"F axis position {self.f_position} close to software limits, defined in the config file.")
                self.Focus_Position_Indicator.setStyleSheet("color: red;")
            else:
                self.Focus_Position_Indicator.setStyleSheet("color: white;")
        self.sig_move_relative.emit(pos_dict)

    @QtCore.pyqtSlot(int)
    def set_laser_intensity(self, value):
        self.sig_state_request.emit({'intensity': value})
        self.LaserIntensitySlider.setValue(value)
        self.LaserIntensitySpinBox.setValue(value)

    def connect_widget_to_state_parameter(self, widget, state_parameter, conversion_factor):
        '''
        Helper method to (currently) connect spinboxes
        '''
        if isinstance(widget,(QtWidgets.QSpinBox, QtWidgets.QDoubleSpinBox)):
            self.connect_spinbox_to_state_parameter(widget, state_parameter, conversion_factor)

    def connect_combobox_to_state_parameter(self, combobox, option_list, state_parameter, int_conversion = False):
        '''
        Helper method to connect and initialize a combobox from the config

        Args:
            combobox (QtWidgets.QComboBox): Combobox in the GUI to be connected
            option_list (list): List of selection options
            state_parameter (str): State parameter (has to exist in the config)
        '''
        combobox.addItems(option_list)
        if not int_conversion:
            combobox.currentTextChanged.connect(lambda currentText: self.sig_state_request.emit({state_parameter : currentText}))
            combobox.setCurrentText(self.cfg.startup[state_parameter])
        else:
            combobox.currentTextChanged.connect(lambda currentParameter: self.sig_state_request.emit({state_parameter : int(currentParameter)}))
            combobox.setCurrentText(str(self.cfg.startup[state_parameter]))

    def connect_spinbox_to_state_parameter(self, spinbox, state_parameter, conversion_factor=1):
        '''
        Helper method to connect and initialize a spinbox from the config

        Args:
            spinbox (QtWidgets.QSpinBox or QtWidgets.QDoubleSpinbox): Spinbox in
                    the GUI to be connected
            state_parameter (str): State parameter (has to exist in the config)
            conversion_factor (float): Conversion factor. If the config is in
                                       seconds, the spinbox displays ms:
                                       conversion_factor = 1000. If the config is
                                       in seconds and the spinbox displays
                                       microseconds: conversion_factor = 1000000
        '''
        spinbox.valueChanged.connect(lambda currentValue: self.sig_state_request.emit({state_parameter : currentValue/conversion_factor}))
        spinbox.setValue(self.cfg.startup[state_parameter]*conversion_factor)

    @QtCore.pyqtSlot(str)
    def execute_script(self, script):
        self.sig_execute_script.emit(script)

    def block_signals_from_controls(self, bool):
        '''
        Helper method to allow blocking of signals from all kinds of controls.

        Needs a list in self.widgets_to_block which has to be created during 
        mesoSPIM_MainWindow.__init__()
        
        Args:
            bool (bool): True if widgets are supposed to be blocked, False if unblocking is desired.
        '''
        for widget in self.widgets_to_block:
            widget.blockSignals(bool)

    def update_widget_from_state(self, widget, state_parameter_string, conversion_factor):
        if isinstance(widget, QtWidgets.QComboBox):
            widget.setCurrentText(self.state[state_parameter_string])
        elif isinstance(widget, (QtWidgets.QSlider,QtWidgets.QDoubleSpinBox,QtWidgets.QSpinBox)):
            widget.setValue(self.state[state_parameter_string]*conversion_factor)
    
    @QtCore.pyqtSlot()
    def update_gui_from_state(self):
        '''
        Updates the GUI controls after a state_change
        if the self.update_gui_from_state_flag is enabled.
        '''
        if self.update_gui_from_state_flag:
            self.block_signals_from_controls(True)
            for widget, state_parameter, conversion_factor in self.widget_to_state_parameter_assignment:
                self.update_widget_from_state(widget, state_parameter, conversion_factor)                
            self.block_signals_from_controls(False)

    def run_snap(self):
        self.sig_state_request.emit({'state':'snap'})
        self.set_progressbars_to_busy()
        self.enable_mode_control_buttons(False)
        self.enable_stop_button(True)
        
    def run_live(self):
        self.sig_state_request.emit({'state':'live'})
        logger.debug('Thread ID during live: '+str(int(QtCore.QThread.currentThreadId())))
        self.sig_poke_demo_thread.emit()
        self.set_progressbars_to_busy()
        self.enable_mode_control_buttons(False)
        self.enable_stop_button(True)

    def run_selected_acquisition(self):
        row = self.acquisition_manager_window.get_first_selected_row()

        if row is None:
            self.display_warning('No row selected - stopping!')
        else:
            # print('selected row:', row)
            self.state['selected_row'] = row
            self.sig_state_request.emit({'state':'run_selected_acquisition'})
            self.enable_mode_control_buttons(False)
            self.enable_gui_updates_from_state(True)
            self.enable_stop_button(True)
            self.enable_gui(False)
            ''' Disabled taskbar button progress display due to problems with Anaconda default
            if sys.platform == 'win32':
                self.win_taskbar_button.progress().setVisible(True)
            '''

    def run_acquisition_list(self):
        self.state['selected_row'] = -1
        self.sig_state_request.emit({'state':'run_acquisition_list'})
        self.enable_mode_control_buttons(False)
        self.enable_gui_updates_from_state(True)
        self.enable_stop_button(True)
        self.enable_gui(False)
        ''' Disabled taskbar button progress display due to problems with Anaconda default
        if sys.platform == 'win32':
            self.win_taskbar_button.progress().setVisible(True)
        '''

    def run_lightsheet_alignment_mode(self):
        self.sig_state_request.emit({'state':'lightsheet_alignment_mode'})
        self.set_progressbars_to_busy()
        self.enable_mode_control_buttons(False)
        self.enable_stop_button(True)
        ''' Disabled taskbar button progress display due to problems with Anaconda default
        if sys.platform == 'win32':
            self.win_taskbar_button.progress().setVisible(False)
        '''
    
    def run_visual_mode(self):
        self.sig_state_request.emit({'state':'visual_mode'})
        self.set_progressbars_to_busy()
        self.enable_mode_control_buttons(False)
        self.enable_stop_button(True)
        ''' Disabled taskbar button progress display due to problems with Anaconda default
        if sys.platform == 'win32':
            self.win_taskbar_button.progress().setVisible(False)
        '''

    @QtCore.pyqtSlot(dict)
    def launch_optimizer(self, ini_dict=None):
        self.sig_move_relative.emit({'f_rel': 5})  # a hack to fix Galil stage coupling, between F and X/Y stages.
        time.sleep(0.1)
        self.sig_move_relative.emit({'f_rel': -5})  # a hack to fix Galil stage coupling, between F and X/Y stages.
        if not self.optimizer:
            self.optimizer = mesoSPIM_Optimizer(self)
            self.optimizer.set_parameters(ini_dict)
        else:
            self.optimizer.set_parameters(ini_dict)
            self.optimizer.show()

    @QtCore.pyqtSlot()
    def launch_contrast_window(self):
        if not self.contrast_window:
            self.contrast_window = mesoSPIM_ContrastWindow(self)
            self.core.camera_worker.sig_camera_frame.connect(self.contrast_window.set_image)
        else:
            self.contrast_window.active = True
            self.contrast_window.show()

    @QtCore.pyqtSlot(bool)
    def enable_gui_updates_from_state(self, boolean):
        self.update_gui_from_state_flag = boolean

    def enable_stop_button(self, boolean):
        self.StopButton.setEnabled(boolean)
    
    def enable_gui(self, boolean):
        self.TabWidget.setEnabled(boolean)
        self.ControlGroupBox.setEnabled(boolean)
        self.sig_enable_gui.emit(boolean)

    def enable_mode_control_buttons(self, boolean):
        self.LiveButton.setEnabled(boolean)
        self.SnapButton.setEnabled(boolean)
        self.RunSelectedAcquisitionButton.setEnabled(boolean)
        self.RunAcquisitionListButton.setEnabled(boolean)
        self.VisualModeButton.setEnabled(boolean)
        self.LightsheetSwitchingModeButton.setEnabled(boolean)

    def finished(self):
        self.enable_gui_updates_from_state(False)
        self.enable_stop_button(False)
        self.enable_mode_control_buttons(True)
        self.enable_gui(True)
        self.set_progressbars_to_standard()

    def set_progressbars_to_busy(self):
        '''If min and max of a progress bar are 0, it shows a "busy" indicator'''
        self.AcquisitionProgressBar.setMinimum(0)
        self.AcquisitionProgressBar.setMaximum(0)
        self.TotalProgressBar.setMinimum(0)
        self.TotalProgressBar.setMaximum(0)
        ''' Disabled taskbar button progress display due to problems with Anaconda default
        if sys.platform == 'win32':
            self.win_taskbar_button.progress().setVisible(False)
        '''
    
    def set_progressbars_to_standard(self):
        self.AcquisitionProgressBar.setMinimum(0)
        self.AcquisitionProgressBar.setMaximum(100)
        self.TotalProgressBar.setMinimum(0)
        self.TotalProgressBar.setMaximum(100)
        ''' Disabled taskbar button progress display due to problems with Anaconda default
        if sys.platform == 'win32':
            self.win_taskbar_button.progress().setValue(0)
            self.win_taskbar_button.progress().setVisible(False)
        '''

    def update_etl_increments(self):
        increment = self.ETLIncrementSpinBox.value()

        self.LeftETLOffsetSpinBox.setSingleStep(increment)
        self.RightETLOffsetSpinBox.setSingleStep(increment)
        self.LeftETLAmplitudeSpinBox.setSingleStep(increment)
        self.RightETLAmplitudeSpinBox.setSingleStep(increment)
   
    def zero_left_etl(self):
        ''' Zeros the amplitude of the left ETL for faster alignment '''
        if self.ZeroLeftETLButton.isChecked():
            self.ETL_L_amp_backup = self.LeftETLAmplitudeSpinBox.value()
            self.LeftETLAmplitudeSpinBox.setValue(0)
        else:
            self.LeftETLAmplitudeSpinBox.setValue(self.ETL_L_amp_backup)

    def zero_right_etl(self):
        ''' Zeros the amplitude of the right ETL for faster alignment '''
        if self.ZeroRightETLButton.isChecked():
            self.ETL_R_amp_backup = self.RightETLAmplitudeSpinBox.value()
            self.RightETLAmplitudeSpinBox.setValue(0)
        else:
            self.RightETLAmplitudeSpinBox.setValue(self.ETL_R_amp_backup)

    def zero_galvo_amp(self):
        '''Set the amplitude of both galvos to zero, or back to where it was, depending on button state'''
        if self.freezeGalvoButton.isChecked():
            self.galvo_amp_backup = self.LeftGalvoAmplitudeSpinBox.value()
            self.LeftGalvoAmplitudeSpinBox.setValue(0)
            self.freezeGalvoButton.setText('Unfreeze galvos')
        else:
            self.LeftGalvoAmplitudeSpinBox.setValue(self.galvo_amp_backup)
            self.freezeGalvoButton.setText('Freeze galvos')

    def choose_etl_config(self):
        ''' File dialog for choosing the config file

        TODO: Check that this is really a .csv-File
        '''
        path , _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Open csv File', self.state['ETL_cfg_file'])

        ''' To avoid crashes, only set the cfg file when a file has been selected:'''
        if path:
            self.state['ETL_cfg_file'] = path
            self.ETLconfigIndicator.setText(path)

            logger.info(f'Main Window: Chose ETL Config File: {path}')

            self.sig_state_request.emit({'ETL_cfg_file' : path})

    def save_etl_config(self):
        ''' Save current ETL parameters into config '''
        self.sig_save_etl_config.emit()

    def display_warning(self, string):
        warning = QtWidgets.QMessageBox.warning(None,'mesoSPIM Warning',
                string, QtWidgets.QMessageBox.Ok)

    def choose_snap_folder(self):
        path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Open csv File', self.state['snap_folder'])
        if path:
            self.state['snap_folder'] = path
            self.SnapFolderIndicator.setText(path)
            print('Chosen Snap Folder:', path)

            #self.sig_state_request.emit({'ETL_cfg_file' : path})
    
    def cascade_all_windows(self):
        if hasattr(self.cfg, 'ui_options') and 'window_pos' in self.cfg.ui_options.keys():
            window_pos = self.cfg.ui_options['window_pos']
        else:
            window_pos = (100, 100)
        self.move(window_pos[0], window_pos[1])
        self.camera_window.move(window_pos[0] + 100, window_pos[1] + 100)
        self.acquisition_manager_window.move(window_pos[0] + 200, window_pos[1] + 200)
        if self.webcam_window:
            self.webcam_window.move(window_pos[0] + 300, window_pos[1] + 300)



'''
Logitech Joystick Classes

pywinusb.hid spawns another thread for joystick event handling which might cause
problems.

Because the signals emitted can only be processed when a QEventLoop is running,
you need something with an eventloop (e.g. a QApplication) even for testing.
'''
from PyQt5 import QtCore

class FarmSimulatorSidePanel(QtCore.QObject):
    '''


    The joystick is set up using the pyqinusb package by using an HidDeviceFilter
    for the side panel values.

    Axis numbers are 0-indexed as per Python convention, i.e. the 6 axes are
    designated "0" to "5".

    Signals:
        sig_button_pressed = QtCore.pyqtSignal(int) # <-- allows handling of buttons
        sig_axis_moved = QtCore.pyqtSignal(int, int) # <-- axis, value
        sig_mode_changed = QtCore.pyqtSignal(str) # <-- Modal switching (XY/ZF mode)
        sig_start_timer =  QtCore.pyqtSignal(int) # <-- timer id, value to emit
        sig_stop_timer = QtCore.pyqtSignal(int) # <-- timer id

    Attributes:
        mode (str): Joysticks can have different modes (e.g. whether analog axes 0-2
                    or 3-5 are selected). This is represented in this attribute.

    '''
    sig_button_pressed = QtCore.pyqtSignal(int) # <-- allows handling of buttons
    sig_axis_moved = QtCore.pyqtSignal(int, int) # <-- axis, value
    sig_mode_changed = QtCore.pyqtSignal(str) # <-- Modal switching (XY/ZF mode)
    sig_start_timer =  QtCore.pyqtSignal(int) # <-- timer id, value to emit
    sig_stop_timer = QtCore.pyqtSignal(int) # <-- timer id

    def __init__(self):
        super().__init__()

        import pywinusb.hid as hid

        self.hid_filter = hid.HidDeviceFilter(vendor_id = 0x0738, product_id = 0x2218)
        self.hid_device = self.hid_filter.get_devices()
        self.joystick = self.hid_device[0]

        self.joystick.open()
        self.joystick.set_raw_data_handler(self.farm_panel_handler)

        self.mode = 'undefined'
        self.sig_mode_changed.emit(self.mode)

        self.sig_start_timer.connect(self.start_axis_timer)
        self.sig_stop_timer.connect(self.stop_axis_timer)

        '''
        One problem with the joystick is that it stops sending packages when the
        maximum tip/tilt is reached. To circumvent the motion to be stopped,
        a QTimer is used to periodically trigger movement in the same direction.

        joystick_timer_start/stop are helper methods.
        '''
        self.timeout_interval = 10
        self.axis0_timer = QtCore.QTimer(self)
        self.axis1_timer = QtCore.QTimer(self)
        self.axis2_timer = QtCore.QTimer(self)
        self.axis3_timer = QtCore.QTimer(self)
        self.axis4_timer = QtCore.QTimer(self)
        self.axis5_timer = QtCore.QTimer(self)

        self.axis0_value = 0
        self.axis1_value = 0
        self.axis2_value = 0
        self.axis3_value = 0
        self.axis4_value = 0
        self.axis5_value = 0

        self.axis0_timer.timeout.connect(lambda: self.sig_axis_moved.emit(0, self.axis0_value))
        self.axis1_timer.timeout.connect(lambda: self.sig_axis_moved.emit(1, self.axis1_value))
        self.axis2_timer.timeout.connect(lambda: self.sig_axis_moved.emit(2, self.axis2_value))
        self.axis3_timer.timeout.connect(lambda: self.sig_axis_moved.emit(3, self.axis3_value))
        self.axis4_timer.timeout.connect(lambda: self.sig_axis_moved.emit(4, self.axis4_value))
        self.axis5_timer.timeout.connect(lambda: self.sig_axis_moved.emit(5, self.axis5_value))

    def start_axis_timer(self, axis):
        value = exec('self.axis'+str(axis)+'_value')
        exec('self.axis'+str(axis)+'_timer.start(self.timeout_interval)')

    def stop_axis_timer(self, axis):
        exec('self.axis'+str(axis)+'_timer.stop()')

    def __del__(self):
        try:
            self.joystick.close()
        except:
            print('Closing HID device failed')

    def sample_handler(self, data):
        print("Raw data: {0}".format(data))
        print(data[1])

    def get_bin(self, x, n=0):
        '''
        Get the binary representation of x.

        Args:
            x (int): Data
            n (int): Minimum number of digits. If x needs less digits in binary, the rest
            is filled with zeros.

        Returns
        -------
        str
        '''
        return format(x, 'b').zfill(n)

    def farm_panel_handler(self, data):
        '''Buttons 1 to 8'''
        self.group_1to8 = data[1]

        self.group_1to8_string = self.get_bin(self.group_1to8,8)
        '''Catch only events which are different from Off-events'''
        if self.group_1to8_string != '00000000':
            button = 8 - self.group_1to8_string.find('1')
            self.sig_button_pressed.emit(button)

        self.group_9to16 = data[2]
        self.group_9to16_string = self.get_bin(self.group_9to16,8)
        '''Catch only events which are different from Off-events'''
        if self.group_9to16_string != '00000000':
            button = 16 - self.group_9to16_string.find('1')
            self.sig_button_pressed.emit(button)

        self.group_17to24 = data[3]
        self.group_17to24_string = self.get_bin(self.group_17to24,8)
        '''Catch only events which are different from Off-events'''
        if self.group_17to24_string != '00000000':
            button = 24 - self.group_17to24_string.find('1')
            self.sig_button_pressed.emit(button)

        self.group_25to29 = data[4]
        self.group_25to29_string = self.get_bin(self.group_25to29,8)
        '''Catch only events which are different from Off-events'''
        if self.group_25to29_string != '00000000':
            index = self.group_25to29_string.find('1')
            if index == 0:
                '''
                29 is the mode changing button, so the corresponding
                signal should be emitted as well:
                '''
                print('self mode: ', self.mode)
                if self.mode == '012':
                    self.mode = '345'
                    self.sig_mode_changed.emit(self.mode)
                elif self.mode == '345':
                    self.mode = '012'
                    self.sig_mode_changed.emit(self.mode)

                self.sig_button_pressed.emit(29)
            elif index == 4:
                self.sig_button_pressed.emit(28)
            else:
                button = 32-index
                self.sig_button_pressed.emit(button)

        '''Joystick handling:

        Stop the joystick timer - a QTimer can be stopped even though it was never
        started. This allows every new arriving HID package to stop the
        persistent sending of messages.
        '''
        self.handle_axis_value_changes(0,'012',5,data)
        self.handle_axis_value_changes(1,'012',6,data)
        self.handle_axis_value_changes(2,'012',7,data)
        self.handle_axis_value_changes(3,'345',8,data)
        self.handle_axis_value_changes(4,'345',9,data)
        self.handle_axis_value_changes(5,'345',10,data)

    def handle_axis_value_changes(self, axis_id, axis_group, data_group, data):
        value = data[data_group]
        if value != 128:
            if self.mode != axis_group:
                self.mode = axis_group
                self.sig_mode_changed.emit(axis_group)

            if value-128 == -128 or value-128 == 127:
                ''' Assign a certain axis the min or max value '''
                exec('self.axis'+str(axis_id)+'_value = value')
                ''' Start timers. Because this is executed from
                another thread, a signal has to be used here.'''
                self.sig_start_timer.emit(axis_id)
                self.sig_axis_moved.emit(axis_id, value)
            else:
                self.sig_stop_timer.emit(axis_id)
                self.sig_axis_moved.emit(axis_id, value)



'''
mesoSPIM Acquisition Manager Window
===================================
'''
import os
import sys
import time

import logging
logger = logging.getLogger(__name__)

from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.uic import loadUi

''' mesoSPIM imports '''
from .mesoSPIM_State import mesoSPIM_StateSingleton

from .utils.models import AcquisitionModel

from .utils.delegates import (ComboDelegate,
                        SliderDelegate,
                        MarkXPositionDelegate,
                        MarkYPositionDelegate,
                        MarkZPositionDelegate,
                        MarkFocusPositionDelegate,
                        ProgressBarDelegate,
                        ZstepSpinBoxDelegate,
                        IntensitySpinBoxDelegate,
                        ChooseFolderDelegate,
                        ETLSpinBoxDelegate,
                        RotationSpinBoxDelegate)

from .utils.widgets import MarkPositionWidget

from .utils.multicolor_acquisition_wizard import MulticolorTilingWizard
from .utils.filename_wizard import FilenameWizard
from .utils.focus_tracking_wizard import FocusTrackingWizard
from .utils.image_processing_wizard import ImageProcessingWizard

from .utils.utility_functions import convert_seconds_to_string

class MyStyle(QtWidgets.QProxyStyle):
    def drawPrimitive(self, element, option, painter, widget=None):
        '''
        Draw a line across the entire row rather than just the column
        we're hovering over.
        '''
        if element == self.PE_IndicatorItemViewItemDrop and not option.rect.isNull():
            option_new = QtWidgets.QStyleOption(option)
            option_new.rect.setLeft(0)
            if widget:
                option_new.rect.setRight(widget.width())
            option = option_new
        super().drawPrimitive(element, option, painter, widget)

class mesoSPIM_AcquisitionManagerWindow(QtWidgets.QWidget):

    model_changed = QtCore.pyqtSignal(AcquisitionModel)
    sig_warning = QtCore.pyqtSignal(str)
    sig_move_absolute = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super().__init__()

        self.parent = parent # mesoSPIM_MainWindow instance
        self.cfg = parent.cfg

        self.state = mesoSPIM_StateSingleton()

        loadUi(self.parent.package_directory + '/gui/mesoSPIM_AcquisitionManagerWindow.ui', self)
        self.setWindowTitle('mesoSPIM Acquisition Manager')

        self.MoveUpButton.clicked.connect(self.move_selected_row_up)
        self.MoveDownButton.clicked.connect(self.move_selected_row_down)

        ''' Parent Enable / Disable GUI'''
        self.parent.sig_enable_gui.connect(lambda boolean: self.setEnabled(boolean))

        self.statusBar = QtWidgets.QStatusBar()

        ''' Setting the model up '''
        self.model = AcquisitionModel()

        self.table.setModel(self.model)
        self.model.dataChanged.connect(self.set_state)
        self.model.dataChanged.connect(self.update_acquisition_time_prediction)

        ''' Table selection behavior '''
        self.table.setSelectionBehavior(self.table.SelectRows)
        self.table.setSelectionMode(self.table.ExtendedSelection)
        self.table.setDragDropMode(self.table.InternalMove)
        self.table.setDragDropOverwriteMode(False)
        self.table.setDragEnabled(True)
        self.table.setAcceptDrops(True)
        self.table.setDropIndicatorShown(True)
        self.table.setSortingEnabled(True)

    
        self.set_item_delegates()

        ''' Set our custom style - this draws the drop indicator across the whole row '''
        self.table.setStyle(MyStyle())

        self.selection_model = self.table.selectionModel()
        self.selection_mapper = QtWidgets.QDataWidgetMapper()

        self.AddButton.clicked.connect(self.add_row)
        self.DeleteButton.clicked.connect(self.delete_row)
        self.CopyButton.clicked.connect(self.copy_row)

        self.SaveButton.clicked.connect(self.save_table)
        self.LoadButton.clicked.connect(self.load_table)

        self.MarkCurrentXYButton.clicked.connect(self.mark_current_xy_position)
        self.MarkCurrentFocusButton.clicked.connect(self.mark_current_focus)
        self.MarkCurrentRotationButton.clicked.connect(self.mark_current_rotation)
        self.MarkCurrentStateButton.clicked.connect(self.mark_current_state)
        self.MarkCurrentETLParametersButton.clicked.connect(self.mark_current_etl_parameters)
        self.MarkAllButton.clicked.connect(self.mark_all_current_parameters)
        self.PreviewSelectionButton.clicked.connect(self.preview_acquisition)

        self.TilingWizardButton.clicked.connect(self.run_tiling_wizard)
        self.FilenameWizardButton.clicked.connect(self.generate_filenames)
        self.FocusTrackingWizardButton.clicked.connect(self.run_focus_tracking_wizard)
        self.ImageProcessingWizardButton.clicked.connect(self.run_image_processing_wizard)

        self.DeleteAllButton.clicked.connect(self.delete_all_rows)
        # self.SetRotationPointButton.clicked.connect(lambda bool: self.set_rotation_point() if bool is True else self.delete_rotation_point())
        self.SetFoldersButton.clicked.connect(self.set_folder_names)

        font = QtGui.QFont()
        font.setPointSize(14)
        self.table.horizontalHeader().setFont(font)
        self.table.verticalHeader().setFont(font)

        logger.info('Thread ID at Startup: '+str(int(QtCore.QThread.currentThreadId())))

        self.selection_model.selectionChanged.connect(self.selected_row_changed)

        ''' Display initial time prediction '''
        self.update_acquisition_time_prediction()

        '''XML writing testcode'''
        # self.GenerateXMLButton.clicked.connect(self.generate_xml)
 

    def enable(self):
        self.setEnabled(True)

    def disable(self):
        self.setEnabled(False)

    def display_status_message(self, string, time=0):
        '''
        Displays a message in the status bar for a time in ms

        If time=0, the message will stay.
        '''

        if time == 0:
            self.statusBar.showMessage(string)
        else:
            self.statusBar.showMessage(string, time)

    def get_first_selected_row(self):
        ''' Little helper method to provide the first row out of a selection range '''
        try:
            indices = self.selection_model.selectedIndexes()
            rows = self.selection_model.selectedRows()
            row = indices[0].row()
        except:
            row = None

        return row

    def set_selected_row(self, row):
        ''' Little helper method to allow setting the selected row '''
        index = self.model.createIndex(row,0)
        # self.selection_model.clearCurrentIndex()
        self.selection_model.select(index,QtCore.QItemSelectionModel.ClearAndSelect)

    def start_selected(self):
        ''' Get the selected row and run this (single) row only

        Indices is a list of selected QModelIndex objects, we're only interested
        in the first.
        '''
        self.disable_gui()
        row = self.get_first_selected_row()
        if row is not None:
            self.sig_start_selected.emit(row)
        else:
            self.display_no_row_selected_warning()

    def selected_row_changed(self, new_selection, old_selection):
        if new_selection.indexes() != []:
            new_row = new_selection.indexes()[0].row()
            for column in self.persistent_editor_column_indices:
                self.table.openPersistentEditor(self.model.index(new_row, column))

        if old_selection.indexes() != []:
            old_row = old_selection.indexes()[0].row()
            for column in self.persistent_editor_column_indices:
                self.table.closePersistentEditor(self.model.index(old_row, column))   

    def add_row(self):
        self.model.insertRows(self.model.rowCount(),1)

    def delete_row(self):
        ''' Deletes the selected row '''
        if self.model.rowCount() > 1:
            row = self.get_first_selected_row()
            if row is not None:
                self.model.removeRows(row,1)
            else:
                self.display_no_row_selected_warning()
        else:
            self.display_warning("Can't delete last row!")

    def delete_all_rows(self):
        ''' 
        Displays a warning that this will delete the entire table
        and then proceeds to delete if the user clicks 'Yes'
        '''
        reply = QtWidgets.QMessageBox.warning(self,"mesoSPIM Warning",
                'Do you want to delete the table?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No )
        
        if reply == QtWidgets.QMessageBox.Yes:
            self.model.deleteTable()

    def copy_row(self):
        row = self.get_first_selected_row()
        if row is not None:
            self.model.copyRow(row)
        else:
            self.display_no_row_selected_warning()

    def move_selected_row_up(self):
        row = self.get_first_selected_row()
        if row is not None:
            if row > 0:
                self.model.moveRow(QtCore.QModelIndex(),row,QtCore.QModelIndex(),row-1)
                self.set_selected_row(row-1)
        else:
            self.display_no_row_selected_warning()

    def move_selected_row_down(self):
        row = self.get_first_selected_row()
        if row is not None:
            if row < self.model.rowCount():
                self.model.moveRow(QtCore.QModelIndex(),row,QtCore.QModelIndex(),row+1)
                self.set_selected_row(row+1)
        else:
            self.display_no_row_selected_warning()

    def set_item_delegates(self):
        ''' Several columns should have certain delegates

        If I decide to move colums, the delegates should move with them

        Here, I need the configuration to provide the options for the
        delegates.
        '''
        self.delegate_dict = {'x_pos' :  'MarkXPositionDelegate(self)',
                              'y_pos' : 'MarkYPositionDelegate(self)',
                              'z_start' : 'MarkZPositionDelegate(self)',
                              'z_end' : 'MarkZPositionDelegate(self)',
                              'z_step' : 'ZstepSpinBoxDelegate(self)',
                              'rot' : 'RotationSpinBoxDelegate(self)',
                              'f_start' : 'MarkFocusPositionDelegate(self)',
                              'f_end' : 'MarkFocusPositionDelegate(self)',
                              'filter' : 'ComboDelegate(self,[key for key in self.cfg.filterdict.keys()])',
                              'intensity' : 'IntensitySpinBoxDelegate(self)',
                              'laser' : 'ComboDelegate(self,[key for key in self.cfg.laserdict.keys()])',
                              'zoom' : 'ComboDelegate(self,[key for key in self.cfg.zoomdict.keys()])',
                              'shutterconfig' : 'ComboDelegate(self,[key for key in self.cfg.shutteroptions])',
                              'folder' : 'ChooseFolderDelegate(self)',
                              'etl_l_offset' : 'ETLSpinBoxDelegate(self)',
                              'etl_l_amplitude' : 'ETLSpinBoxDelegate(self)',
                              'etl_r_offset' : 'ETLSpinBoxDelegate(self)',
                              'etl_r_amplitude' : 'ETLSpinBoxDelegate(self)',
                              }

        self.persistent_editor_column_indices=[]
        ''' Go through the dictionary keys of the

        self.model._table[0].

        find the index of a certain key and set the delegate accordingly

        '''
        for key in self.delegate_dict :
            column_index = self.model._table[0].keys().index(key)
            ''' As some of the delegates expect options, a hack using exec was used: '''
            string_to_execute = 'self.table.setItemDelegateForColumn(column_index,'+self.delegate_dict[key]+')'
            delegate_object = exec(self.delegate_dict[key])

            self.persistent_editor_column_indices.append(column_index)
            exec(string_to_execute)

    def update_acquisition_time_prediction(self):
        framerate = self.state['current_framerate']
        total_time = self.state['acq_list'].get_acquisition_time(framerate)
        self.state['predicted_acq_list_time'] = total_time
        self.state['remaining_acq_list_time'] = total_time
        time_string = convert_seconds_to_string(total_time)
        self.AcquisitionTimeEdit.setText(time_string)

    def set_state(self):
        self.state['acq_list'] = self.model.get_acquisition_list()
        
    def enable_gui(self):
        ''' Enables all GUI controls, disables stop button '''
        self.table.setEnabled(True)
        self.tableControlButtons.setEnabled(True)
        self.generalControlButtons.setEnabled(True)

    def disable_gui(self):
        ''' Disables all buttons and controls, enables stop button '''
        self.table.setEnabled(False)
        self.tableControlButtons.setEnabled(False)
        self.generalControlButtons.setEnabled(False)

    def save_table(self):
        path , _ = QtWidgets.QFileDialog.getSaveFileName(None,'Save Table')
        if path:
            self.model.saveModel(path)
        self.set_state()

    def load_table(self):
        path , _ = QtWidgets.QFileDialog.getOpenFileName(None,'Load Table')
        if path:
            try:
                self.model.loadModel(path)
                self.set_state()
                self.update_acquisition_time_prediction()
            except:
                self.sig_warning.emit('Table cannot be loaded - incompatible file format (Probably created by a previous version of the mesoSPIM software)!')

    def run_tiling_wizard(self):
        wizard = MulticolorTilingWizard(self)

    def run_focus_tracking_wizard(self):
        wizard = FocusTrackingWizard(self)

    def run_image_processing_wizard(self):
        wizard = ImageProcessingWizard(self)

    def mark_current_xy_position(self):
        row = self.get_first_selected_row()

        if row is not None:
            self.model.setDataFromState(row, 'x_pos')
            self.model.setDataFromState(row, 'y_pos')
        else:
            if self.model.rowCount() == 1:
                self.set_selected_row(0)
                self.mark_current_xy_position()
            else:
                self.display_no_row_selected_warning()

    def mark_current_state(self):
        row = self.get_first_selected_row()

        if row is not None:
            self.model.setDataFromState(row, 'filter')
            self.model.setDataFromState(row, 'zoom')
            self.model.setDataFromState(row, 'laser')
            self.model.setDataFromState(row, 'intensity')
            self.model.setDataFromState(row, 'shutterconfig')
        else:
            if self.model.rowCount() == 1:
                self.set_selected_row(0)
                self.mark_current_state()
            else:
                self.display_no_row_selected_warning()

    def mark_current_etl_parameters(self):
        row = self.get_first_selected_row()

        if row is not None:
            self.model.setDataFromState(row, 'etl_l_offset')
            self.model.setDataFromState(row, 'etl_l_amplitude')
            self.model.setDataFromState(row, 'etl_r_offset')
            self.model.setDataFromState(row, 'etl_r_amplitude')
        else:
            if self.model.rowCount() == 1:
                self.set_selected_row(0)
                self.mark_current_etl_parameters()
            else:
                self.display_no_row_selected_warning()

    def mark_current_focus(self):
        ''' Marks both foci start focus '''
        row = self.get_first_selected_row()

        if row is not None:
            f_pos = self.state['position']['f_pos']
            ''' Set f_start and f_end to the same values '''
            column_index0 = self.model._table[0].keys().index('f_start')
            index0 = self.model.createIndex(row, column_index0)
            column_index1 = self.model._table[0].keys().index('f_end')
            index1 = self.model.createIndex(row, column_index1)

            self.model.setData(index0, f_pos)
            self.model.setData(index1, f_pos)
        
        else:
            if self.model.rowCount() == 1:
                self.set_selected_row(0)
                self.mark_current_focus()
            else:
                self.display_no_row_selected_warning()
            
    def mark_current_rotation(self):
        row = self.get_first_selected_row()

        if row is not None:
            self.model.setDataFromState(row, 'rot')

        else:
            if self.model.rowCount() == 1:
                self.set_selected_row(0)
                self.mark_current_rotation()
            else:
                self.display_no_row_selected_warning()

    def mark_all_current_parameters(self):
        self.mark_current_xy_position()
        self.mark_current_rotation()
        self.mark_current_focus()
        self.mark_current_etl_parameters()
        self.mark_current_state()

    def preview_acquisition(self):
        row = self.get_first_selected_row()
        # print('selected row:', row)
        if row is not None:
            self.state['selected_row'] = row
            # Check if the z position should be updated
            if self.PreviewZCheckBox.checkState():
                # print('Checkbox checked')
                self.parent.sig_state_request.emit({'state':'preview_acquisition_with_z_update'})
            else:
                # print('Checkbox not checked')
                self.parent.sig_state_request.emit({'state':'preview_acquisition_without_z_update'})
        else: 
            if self.model.rowCount() == 1:
                self.set_selected_row(0)
                self.preview_acquisition() # recursive call of the same function
            else:    
                self.display_no_row_selected_warning()

    def set_folder_names(self):
        path = QtWidgets.QFileDialog.getExistingDirectory(self.parent, 'Select Folder')
        if path:
            column_index = self.model._table[0].keys().index('folder')
            for row in range(0, self.model.rowCount()):
                index = self.model.createIndex(row, column_index)
                self.model.setData(index, path)

    def generate_filenames(self):
        wizard = FilenameWizard(self)

    def display_no_row_selected_warning(self):
        self.display_warning('No row selected!')

    def display_warning(self, string):
        warning = QtWidgets.QMessageBox.warning(None,'mesoSPIM Warning',
                string, QtWidgets.QMessageBox.Ok) 







'''
Simple webcam display widget, can be used stand-alone
author: Nikita Vladimirov, @nvladimus, 2022
License: GPL-3
'''

import numpy as np
import pyqtgraph as pg
import sys

import logging
logger = logging.getLogger(__name__)

from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.uic import loadUi
from PyQt5.QtMultimedia import QCameraInfo, QCamera, QCameraViewfinderSettings
from PyQt5.QtMultimediaWidgets import QCameraViewfinder


class WebcamWindow(QtWidgets.QWidget):
    sig_state_request = QtCore.pyqtSignal(dict)
    sig_move_absolute = QtCore.pyqtSignal(dict)

    def __init__(self, webcam_id=0):
        super().__init__()
        #self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.WEBCAM_ID = webcam_id
        loadUi('gui/WebcamWindow.ui', self)
        self.setWindowTitle(f'Webcam view, camera ID {self.WEBCAM_ID}')
        self.show()
        self.start_capture()

    def start_capture(self):
        webcams = QCameraInfo.availableCameras()
        print(f"Webcams found: {len(webcams)}")
        if len(webcams) > 0:
            self.webcam = QCamera(webcams[self.WEBCAM_ID])
            self.webcam.setCaptureMode(QCamera.CaptureViewfinder)
            self.vf_settings = QCameraViewfinderSettings()
            #self.vf_settings.setResolution(960, 720)
            self.webcam.setViewfinderSettings(self.vf_settings)
            self.webcam.setViewfinder(self.viewfinder)
            #self.viewfinder = QCameraViewfinder() # this object is defined inside the WebcamWindow.ui file
            self.webcam.start()
            #print(f"Supported webcam viewfinder resolutions {self.webcam.supportedViewfinderResolutions()}")
            self.viewfinder.show()
        else:
            pass


if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    window = WebcamWindow()
    sys.exit(app.exec_())


'''
mesoSPIM CameraWindow
'''
import sys
import numpy as np
from .utils.optimization import shannon_dct

import logging
logger = logging.getLogger(__name__)

from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.uic import loadUi
import pyqtgraph as pg
from .mesoSPIM_State import mesoSPIM_StateSingleton

class mesoSPIM_CameraWindow(QtWidgets.QWidget):
    sig_update_roi = QtCore.pyqtSignal(tuple)
    sig_update_status = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__()

        self.parent = parent # the mesoSPIM_MainWindow() instance
        self.cfg = parent.cfg
        self.state = mesoSPIM_StateSingleton()

        pg.setConfigOptions(imageAxisOrder='row-major')
        if (hasattr(self.cfg, 'ui_options') and self.cfg.ui_options['dark_mode']) or\
                (hasattr(self.cfg, 'dark_mode') and self.cfg.dark_mode):
            pg.setConfigOptions(background=pg.mkColor('#19232D'))  # To avoid pitch black bg for the image view
        else:
            pg.setConfigOptions(background="w")

        '''Set up the UI'''
        if __name__ == '__main__':
            loadUi('../gui/mesoSPIM_CameraWindow.ui', self)
        else:
            loadUi(self.parent.package_directory + '/gui/mesoSPIM_CameraWindow.ui', self)
        self.setWindowTitle('mesoSPIM-Control: Camera Window')

        self.status_label.setText("Status: OK")

        ''' Set histogram Range '''
        self.image_view.setLevels(100, 3000)
        self.imageItem = self.image_view.getImageItem()
        self.histogram = self.image_view.getHistogramWidget()
        self.histogram.setMinimumWidth(100)
        self.histogram.item.vb.setMaximumWidth(100)

        ''' This is flipped to account for image rotation '''
        self.y_image_width = self.cfg.camera_parameters['x_pixels']
        self.x_image_width = self.cfg.camera_parameters['y_pixels']
        self.subsampling = self.cfg.startup['camera_display_live_subsampling']

        ''' Initialize crosshairs '''
        self.crosspen = pg.mkPen({'color': "r", 'width': 1})
        self.vLine = pg.InfiniteLine(pos=self.x_image_width/2, angle=90, movable=False, pen=self.crosspen)
        self.hLine = pg.InfiniteLine(pos=self.y_image_width/2, angle=0, movable=False, pen=self.crosspen)
        self.image_view.addItem(self.vLine, ignoreBounds=True)
        self.image_view.addItem(self.hLine, ignoreBounds=True)

        # Create overlay ROIs
        self.overlay = None  # None, 'box'
        w, h = self.x_image_width//self.subsampling, self.y_image_width//self.subsampling
        self.roi_box = pg.RectROI((0, 0), (w, h), sideScalers=True)
        self.roi_drawn = False

        # Set up internal CameraWindow signals
        self.adjustLevelsButton.clicked.connect(self.adjust_levels)
        self.overlayCombo.currentTextChanged.connect(self.change_overlay)
        self.roi_box.sigRegionChangeFinished.connect(self.update_status)
        self.sig_update_status.connect(self.update_status)

        logger.info('Thread ID at Startup: '+str(int(QtCore.QThread.currentThreadId())))

    def adjust_levels(self, pct_low=25, pct_hi=99.99):
        ''''Adjust histogram levels'''
        img = self.image_view.getImageItem().image
        self.image_view.setLevels(min=np.percentile(img, pct_low), max=np.percentile(img, pct_hi))

    def px2um(self, px, scale=1):
        '''Unit converter'''
        return scale * px * self.cfg.pixelsize[self.state['zoom']]

    @QtCore.pyqtSlot(str)
    def change_overlay(self, overlay_name):
        w, h = self.get_image_shape()
        if overlay_name == 'Box roi':
            self.set_roi('box', (w//2 - 50, h//2 - 50, 100, 100))
        elif overlay_name == 'Overlay: none':
            self.set_roi(None, (0, 0, w, h))

    def get_roi(self):
        im_item = self.image_view.getImageItem()
        if self.overlay == 'box' and self.roi_drawn:
            roi = self.roi_box.getArrayRegion(im_item.image, im_item)
            x, y = self.roi_box.pos()
            w, h = self.roi_box.size()
            self.sig_update_roi.emit((x, y, w, h))
        else:
            roi = im_item.image
            w, h = im_item.image.shape
            self.sig_update_roi.emit((0, 0, w, h))
        return roi

    def set_roi(self, mode='box', x_y_w_h=(0, 0, 100, 100)):
        assert mode in ('box', None), f"Mode must be in ('box', None), received {mode} instead"
        self.overlay = mode
        x, y, w, h = x_y_w_h
        self.roi_box.setPos((x, y))
        self.roi_box.setSize((w, h))
        if self.overlay is None and self.roi_drawn:
            self.image_view.removeItem(self.roi_box)
            self.roi_drawn = False
        elif self.overlay == 'box' and not self.roi_drawn:
            self.image_view.addItem(self.roi_box)
            self.roi_drawn = True
        self.sig_update_status.emit()

    def get_image_shape(self):
        return self.image_view.getImageItem().image.shape

    @QtCore.pyqtSlot()
    def update_status(self):
        roi = self.get_roi()
        if self.overlay == 'box':
            w, h = self.roi_box.size()
            self.status_label.setText(f"ROI: w {int(self.px2um(w, self.subsampling)):,} \u03BCm, "
                                      f"h {int(self.px2um(h, self.subsampling)):,} \u03BCm, "
                                      f"sharpness {np.round(1e4 * shannon_dct(roi)):.0f}")
        else:
            self.status_label.setText(f"Image dimensions: {roi.shape}")

    def draw_crosshairs(self):
        self.image_view.addItem(self.vLine, ignoreBounds=True)
        self.image_view.addItem(self.hLine, ignoreBounds=True)

    @QtCore.pyqtSlot(np.ndarray)
    def set_image(self, image):
        self.image_view.setImage(image, autoLevels=False, autoHistogramRange=False, autoRange=False)
        # update roi size if subsampling has changed interactively:
        if self.overlay == 'box' and self.subsampling != self.state['camera_display_live_subsampling']:
            subsampling_ratio = self.subsampling / self.state['camera_display_live_subsampling']
            self.subsampling = self.state['camera_display_live_subsampling']
            x, y = self.roi_box.pos()
            w, h = self.roi_box.size()
            self.roi_box.setPos((x * subsampling_ratio, y * subsampling_ratio))
            self.roi_box.setSize((w * subsampling_ratio, h * subsampling_ratio))
        self.update_status()

        h, w = image.shape[-2], image.shape[-1]  # works for both 2D and 3/4D loaded TIFF files.
        if h != self.y_image_width or w != self.x_image_width:
            self.x_image_width, self.y_image_width = w, h
            self.vLine.setPos(self.x_image_width/2.), self.hLine.setPos(self.y_image_width/2.)
        self.draw_crosshairs()




"""
General-purpose Dynamixel servo class to control e.g. mesoSPIM zoom or filter wheel
Authors: Fabian Voigt, Nikita Vladimirov
"""


import time
from PyQt5 import QtCore
from . import dynamixel_functions as dynamixel_func
import numpy as np
import logging

logger = logging.getLogger(__name__)
PROTOCOL_VERSION = 1


class Dynamixel(QtCore.QObject):
    def __init__(self, COMport, identifier=1, baudrate=115200):
        super().__init__()
        self.dynamixel = dynamixel_func
        self.id = identifier
        self.devicename = COMport.encode('utf-8') # bad naming convention
        self.baudrate = baudrate

        self.addr_mx_mult_turn_offset = 20
        self.addr_mx_torque_enable = 24
        self.addr_mx_goal_position = 30
        self.addr_mx_present_position = 36
        self.addr_mx_p_gain = 28
        self.addr_mx_torque_limit = 34
        self.addr_mx_moving_speed = 32

        ''' Specifies how much the goal position can be off (+/-) from the target '''
        self.goal_position_offset = 10
        self.multiturn_offset = 0
        self.moving_speed = 400
        ''' Specifies how long to sleep for the wait until done function'''
        self.sleeptime = 0.05
        self.timeout = 5

        # the dynamixel library uses integers instead of booleans for binary information
        self.torque_enable = 1
        self.torque_disable = 0
        self._connect()

    def _connect(self):
        try:
            logger.info(f"Connecting to serial port {self.devicename}")
            self.port_num = self.dynamixel.portHandler(self.devicename)
            self.dynamixel.packetHandler()
            self.dynamixel.openPort(self.port_num)
            self.dynamixel.setBaudRate(self.port_num, self.baudrate)
            self.multiturn_offset = self.dynamixel.read2ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_mult_turn_offset)
            self.multiturn_offset = self.normalize_position(self.multiturn_offset)
            logger.info(f"Dynamixel multi-turn offset: {self.multiturn_offset}")
        except Exception as e:
            logger.error(f"Failed to open serial port: {e}")

    def _move(self, position, wait_until_done=False):
        # Enable servo
        self.dynamixel.write1ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_torque_enable, self.torque_enable)
        # Write Moving Speed
        self.dynamixel.write2ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_moving_speed, self.moving_speed)
        # Write Torque Limit
        self.dynamixel.write2ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_torque_limit, 200)
        # Write P Gain
        self.dynamixel.write1ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_p_gain, 44)
        # Write Goal Position
        self.dynamixel.write2ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_goal_position, position)
        # Check position

        ''' This works even though the positions returned during movement are just crap
        - they have 7 to 8 digits. Only when the motor stops, positions are accurate
        -
        '''
        if wait_until_done:
            logger.info(f"Dynamixel (wait_until_done=True) goal positon {position}")
            start_time = time.time()
            upper_limit = position + self.goal_position_offset
            # print('Upper Limit: ', upper_limit)
            lower_limit = position - self.goal_position_offset
            # print('lower_limit: ', lower_limit)
            cur_position = np.int16(self.dynamixel.read2ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_present_position))
            cur_position = self.offset_position(self.normalize_position(cur_position))

            while (cur_position < lower_limit) or (cur_position > upper_limit):
                ''' Timeout '''
                if time.time() - start_time > self.timeout:
                    logger.error("Dynamixel zoom servo: timeout")
                    break
                time.sleep(self.sleeptime)
                cur_position = self.dynamixel.read2ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_present_position)
                cur_position = self.offset_position(self.normalize_position(cur_position))
                logger.info(f"Dynamixel normalized current position {cur_position}")

    def read_position(self):
        '''
        Returns position as an int between 0 and 4096
        '''
        cur_position = self.dynamixel.read4ByteTxRx(self.port_num, PROTOCOL_VERSION, self.id, self.addr_mx_present_position)
        return cur_position

    def normalize_position(self, pos):
        '''Negative numbers representation'''
        if pos >= 65535 // 2:
            cur_position = pos - 65535
        else:
            cur_position = pos
        return cur_position

    def offset_position(self, pos):
        '''Taking into account encoder offset for multi-turn mode.
        The encoder offset can be user-defined eg in Dynamixel Wizard software'''
        if self.multiturn_offset != 0:
            cur_position = pos - self.multiturn_offset
        else:
            cur_position = pos
        return cur_position

    def __del__(self):
        self.dynamixel.closePort(self.port_num)



self.sig_polling_stage_position_stop.emit()

for i in range(100):
	response = self.serial_worker.stage.asi_stages._send_command('W VZRXY\r'.encode('ascii'))
	logger.info(f"Received: {response}")
    
self.sig_polling_stage_position_start.emit()


''' Classes that define acquisition builders:

Take a dict with information and return an acquisition list
'''
from .acquisitions import Acquisition, AcquisitionList


class AcquisitionListBuilder():
    '''
    Generic Acquisition List Builder as parent class?

    TODO: Write this class
    '''
    pass


class MulticolorTilingAcquisitionListBuilder():
    '''
    TODO: Filename generation

    self.dict['x_start']
    self.dict['x_end']
    self.dict['y_start']
    self.dict['y_end']
    self.dict['z_start']
    self.dict['z_end']
    self.dict['z_step']
    self.dict['theta_pos']
    self.dict['f_pos']
    self.dict['x_offset'] # Offset always larger than 0
    self.dict['y_offset'] # Offset always larger than 0
    self.dict['x_image_count']
    self.dict['y_image_count']
    '''

    def __init__(self, dict):
        self.acq_prelist = []

        self.dict = dict

        self.x_start = self.dict['x_start']
        self.y_start = self.dict['y_start']
        self.x_end = self.dict['x_end']
        self.y_end = self.dict['y_end']

        ''' 
        Reverse direction of the offset if pos_end < pos_start
        '''
        if self.x_start < self.x_end:
            self.x_offset = self.dict['x_offset']
        else:
            self.x_offset = -self.dict['x_offset']
            
        if self.y_start < self.y_end:
            self.y_offset = self.dict['y_offset']
        else:
            self.y_offset = -self.dict['y_offset']

        '''
        Core loop: Create an acquisition list for all x & y & channel values
        '''
        tilecount = 0
        n_tiles = self.dict['x_image_count'] * self.dict['y_image_count']
        for i in range(0, self.dict['x_image_count']):
            self.x_pos = round(self.x_start + i * self.x_offset, 2)
            for j in range(0, self.dict['y_image_count']):
                self.y_pos = round(self.y_start + j * self.y_offset, 2)
                channelcount = 0
                for c in range(0, len(self.dict['channels'])):
                    ''' Get a single channeldict out of the list of dicts '''
                    channeldict = self.dict['channels'][c]
                    if self.dict['shutter_seq']:
                        n_shutter_configs = 2
                        shutter_states = ["Left", "Right"]
                    else:
                        n_shutter_configs = 1
                        shutter_states = [self.dict['shutterconfig']]
                    for i_shutter in range(n_shutter_configs):
                        acq = Acquisition(  x_pos=self.x_pos,
                                            y_pos=self.y_pos,
                                            z_start=self.dict['z_start'],
                                            z_end=self.dict['z_end'],
                                            z_step=self.dict['z_step'],
                                            theta_pos=self.dict['theta_pos'],
                                            f_start=round(channeldict['f_start'], 2),
                                            f_end=round(channeldict['f_end'], 2),
                                            laser=channeldict['laser'],
                                            intensity=channeldict['intensity'],
                                            filter=channeldict['filter'],
                                            zoom=self.dict['zoom'],
                                            shutterconfig=shutter_states[i_shutter],
                                            folder=self.dict['folder'],
                                            filename='Tile'+str(tilecount % n_tiles)
                                                     + '_Ch'+str(channelcount)
                                                     + '_Sh' + str(i_shutter % n_shutter_configs),
                                            etl_l_offset=channeldict['etl_l_offset'],
                                            etl_l_amplitude=channeldict['etl_l_amplitude'],
                                            etl_r_offset=channeldict['etl_r_offset'],
                                            etl_r_amplitude=channeldict['etl_r_amplitude'],
                                            )
                        ''' Update number of planes as this is not done by the acquisition 
                        object itself '''
                        acq['planes'] = acq.get_image_count()
                        self.acq_prelist.append(acq)
                    channelcount += 1

                tilecount += 1

    def get_acquisition_list(self):
        return AcquisitionList(self.acq_prelist)



class Demo_Shutter:
    def __init__(self, shutterline):   
        self.shutterline =  shutterline
        self.shutterstate = False

    def open(self, *args):
        self.shutterstate = True

    def close(self, *args):
        self.shutterstate = False
    
    def state(self, *args):
        return self.shutterstate




from PyQt5 import QtWidgets, QtGui, QtCore, QtDesigner

class MarkPositionWidget(QtWidgets.QWidget):
    """ Pushbutton plus position indicator """
    pressed = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        #QtWidgets.QWidget.__init__(self, parent)
        super().__init__(parent)
        # QtDesigner.QPyDesignerCustomWidgetPlugin.__init__(self, parent)

        self.button = QtWidgets.QPushButton()
        self.button.setText("M")

        font = QtGui.QFont()
        font.setPointSize(14)
        self.button.setFont(font)

        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button.sizePolicy().hasHeightForWidth())
        self.button.setSizePolicy(sizePolicy)
        self.button.setMinimumSize(QtCore.QSize(25, 0))

        self.lineEdit = QtWidgets.QLineEdit()
        self.lineEdit.setFont(font)
        # self.lineEdit.setReadOnly(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit.sizePolicy().hasHeightForWidth())
        self.lineEdit.setSizePolicy(sizePolicy)

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.button)
        self.layout.addWidget(self.lineEdit)
        ''' '''
        self.layout.setContentsMargins(0,0,0,0)

        self.button.clicked.connect(lambda: self.pressed.emit())

        self.setLayout(self.layout)
        self.setAutoFillBackground(True)

class SliderWithValue(QtWidgets.QWidget):
    ''' Slider with value to the left '''
    valueChanged = QtCore.pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.label = QtWidgets.QLabel()
        self.label.setMinimumSize(QtCore.QSize(25, 0))

        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)

        self.slider = QtWidgets.QSlider()

        self.slider.setOrientation(QtCore.Qt.Horizontal)
        self.slider.setTracking(False)
        self.slider.valueChanged.connect(self.setText)
        self.slider.valueChanged.connect(self.valueChanged.emit)

        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.slider.sizePolicy().hasHeightForWidth())
        self.slider.setSizePolicy(sizePolicy)


        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.slider)
        self.layout.setContentsMargins(0,0,0,0)

        self.setLayout(self.layout)
        self.setAutoFillBackground(True)

        font = QtGui.QFont()
        font.setPointSize(14)
        self.slider.setFont(font)

    def setText(self, value):
        self.label.setText(str(value) + '%')

    def setValue(self, value):
        self.slider.setValue(value)
        
    def value(self):
        return self.slider.value()



'''
mesoSPIM Filterwheel classes
Authors: Fabian Voigt, Nikita Vladimirov
'''
import time
import serial
import io
from PyQt5 import QtWidgets, QtCore, QtGui
import logging
logger = logging.getLogger(__name__)
from .devices.servos.dynamixel.dynamixel import Dynamixel


class mesoSPIM_DemoFilterWheel(QtCore.QObject):
    def __init__(self, filterdict):
        super().__init__()
        self.filterdict = filterdict

    def _check_if_filter_in_filterdict(self, filter):
        '''
        Checks if the filter designation (string) given as argument
        exists in the filterdict
        '''
        if filter in self.filterdict:
            return True
        else:
            raise ValueError('Filter designation not in the configuration')

    def set_filter(self, filter, wait_until_done=False):
        if self._check_if_filter_in_filterdict(filter) is True:
            if wait_until_done:
                time.sleep(1)


class DynamixelFilterWheel(Dynamixel):
    def __init__(self, filterdict, COMport, identifier=1, baudrate=115200):
        super().__init__(COMport, identifier, baudrate)
        self.filterdict = filterdict
        self.filter = None

    def set_filter(self, filter, wait_until_done=False):
        """Changes filter after checking that the commanded value exists"""
        if filter in self.filterdict:
            self._move(self.filterdict[filter], wait_until_done)
            self.filter = filter
        else:
            raise ValueError('Filter designation not in the configuration')


class LudlFilterWheel(QtCore.QObject):
    """ Class to control a 10-position Ludl filterwheel

    Needs a dictionary which combines filter designations and position IDs
    in the form:

    filters = {'405-488-647-Tripleblock' : 0,
           '405-488-561-640-Quadrupleblock': 1,
           '464 482-35': 2,
           '508 520-35': 3,
           '515LP':4,
           '529 542-27':5,
           '561LP':6,
           '594LP':7,
           'Empty-Alignment':8,}

    If there are tuples instead of integers as values, the
    filterwheel is assumed to be a double wheel.

    I.e.: '508 520-35': (2,3)
    """

    def __init__(self, COMport, filterdict, baudrate=9600):
        super().__init__()

        self.COMport = COMport
        self.baudrate = baudrate
        self.filterdict = filterdict
        self.double_wheel = False
        self.ser = None
        self.sio = None
        self._connect()
        ''' Delay in s for the wait until done function '''
        self.wait_until_done_delay = 0.5

        """
        If the first entry of the filterdict has a tuple
        as value, it is assumed that it is a double-filterwheel
        to change the serial commands accordingly.

        TODO: This doesn't check that the tuple has length 2.
        """
        self.first_item_in_filterdict = list(self.filterdict.keys())[0]

        if type(self.filterdict[self.first_item_in_filterdict]) is tuple:
            self.double_wheel = True

    def _connect(self):
        """"Note: Only one connection should be done per session. Connecting frequently is error-prone,
         because COM port can be scanned by another program (e.g. laser control) and thus be permission-denied at random
          times."""
        try:
            self.ser = serial.Serial(self.COMport,
                                     self.baudrate,
                                     parity=serial.PARITY_NONE,
                                     timeout=0, write_timeout=0,
                                     xonxoff=False,
                                     stopbits=serial.STOPBITS_TWO)
            self.sio = io.TextIOWrapper(io.BufferedRWPair(self.ser, self.ser))
        except serial.SerialException as e:
            logger.error(f"Serial connection to Ludl filter wheel failed: {e}")

    def _check_if_filter_in_filterdict(self, filter):
        '''
        Checks if the filter designation (string) given as argument
        exists in the filterdict
        '''
        if filter in self.filterdict:
            return True
        else:
            raise ValueError('Filter designation not in the configuration')

    def set_filter(self, filter, wait_until_done=False):
        '''
        Moves filter using the pyserial command set.
        No checks are done whether the movement is completed or
        finished in time.
        '''
        if self._check_if_filter_in_filterdict(filter) is True:
            """
            Check for double or single wheel
            TODO: A bit of repeating code in here. Might be better to
            spin the create and send commands off.
            """
            if self.double_wheel is False:
                """ Single wheel code """
                # Get the filter position from the filterdict:
                self.filternumber = self.filterdict[filter]
                # Rotat is the Ludl high-level command for moving a filter wheel
                self.ser.flush()
                self.ludlstring = 'Rotat S M ' + str(self.filternumber) + '\n'
                self.sio.write(str(self.ludlstring))
                self.sio.flush()

                if wait_until_done:
                    ''' Wait a certain number of seconds. This is a hack

                    Testing with :
                    self.sio.write(str('Rdstat S'))
                    self.sio.flush()
                    print('First:', self.sio.readline(10))
                    time.sleep(0.1)
                    self.sio.write(str('Rdstat S'))
                    self.sio.flush()
                    print('Second: ', self.sio.readline(10))

                    yielded very unstable results, sometimes ":N -3", sometimes
                    ":A" - and blocking & crashing the connection
                    '''
                    time.sleep(self.wait_until_done_delay)

            else:
                """ Double wheel code """
                # Get the filter position tuple from the filterdict:
                self.filternumber = self.filterdict[filter]
                """ Write command for the primary wheel """
                self.ludlstring0 = 'Rotat S M ' + str(self.filternumber[0]) + '\n'
                self.sio.write(str(self.ludlstring0))
                """ Write command for the auxillary wheel """
                self.ludlstring1 = 'Rotat S A ' + str(self.filternumber[1]) + '\n'
                self.sio.write(str(self.ludlstring1))
                self.sio.flush()

                if wait_until_done:
                    time.sleep(self.wait_until_done_delay)
        else:
            logger.error(f'Filter {filter} not found in configuration.')

    def __del__(self):
        self.sio.flush()
        self.ser.close()


class SutterLambda10BFilterWheel:
    def __init__(self, comport, filterdict, baudrate=9600, read_on_init=True):
        super().__init__()
        self.COMport = comport
        self.baudrate = baudrate
        self.filterdict = filterdict
        self.double_wheel = False

        ''' Delay in s for the wait until done function '''
        self.wait_until_done_delay = 0.5

        self.first_item_in_filterdict = list(self.filterdict.keys())[0]
        if type(self.filterdict[self.first_item_in_filterdict]) is tuple:
            self.double_wheel = True

        # Open Serial Port
        try:
            self.serial = serial.Serial(self.COMport, self.baudrate, timeout=.25)
        except serial.SerialException:
            raise UserWarning('Could not open the serial port to the Sutter Lambda 10-B.')

        # Place Controller Into Online Mode
        self.serial.write(bytes.fromhex('ee'))

        # Check to see if the initialization sequence has finished.
        if read_on_init:
            self.read(2)  # class 'bytes'
            self.init_finished = True
            logger.info('Done initializing filter wheel')
        else:
            self.init_finished = False
        self.filternumber = 0

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _check_if_filter_in_filterdict(self, filterposition):
        # Checks if the filter designation (string) given as argument exists in the filterdict
        if filterposition in self.filterdict:
            return True
        else:
            raise ValueError('Filter designation not in the configuration')

    def set_filter(self, filterposition=0, speed=0, wait_until_done=False):
        # Confirm that the filter is present in the filter dictionary
        if self._check_if_filter_in_filterdict(filterposition) is True:

            # Confirm that you are only operating in a single filter wheel configuration.
            if self.double_wheel is False:

                # Identify the Filter Number from the Filter Dictionary
                self.wheel_position = self.filterdict[filterposition]

                # Make sure you are moving it to a reasonable filter position, at a reasonable speed.
                assert self.wheel_position in range(10)
                assert speed in range(8)

                # If previously we did not confirm that the initialization was complete, check now.
                if not self.init_finished:
                    self.read(2)
                    self.init_finished = True
                    logger.info('Done initializing filter wheel.')

                # Filter Wheel Command Byte Encoding = wheel + (speed*16) + position = command byte
                outputcommand = self.wheel_position + 16 * speed
                outputcommand = outputcommand.to_bytes(1, 'little')

                # Send out Command
                self.serial.write(outputcommand)
                if wait_until_done:
                    time.sleep(self.wait_until_done_delay)

                # Read up to 2 bytes
                self.read(2)

            else:
                raise UserWarning("Sutter Operates only in a Single Filter Wheel Configuration.")

    def read(self, num_bytes):
        for i in range(100):
            num_waiting = self.serial.inWaiting()
            if num_waiting == num_bytes:
                break
            time.sleep(0.02)
        else:
            raise UserWarning("The serial port to the Sutter Lambda 10-B is on, but it isn't responding as expected.")
        return self.serial.read(num_bytes)

    def close(self):
        self.set_filter()
        self.serial.close()



"""
mesoSPIM Module for controlling Galil-Stages (by Steinmayer-Mechatronik/Feinmess)

Author: Fabian Voigt

#TODO
"""
import copy
import time
import traceback

from PyQt5 import QtWidgets, QtCore, QtGui

from . import gclib

import logging
logger = logging.getLogger(__name__)


class StageControlGalil(QtCore.QObject):
    '''
    Class to control a Galil mechanical stage controller
    
    Inherits from QtCore.QObject so it can be moved to a QThread.
    
    Note:
        The encodercounts should be specfified such that: Position in um = encoderposition * encodercounts
        or Position in degrees = encoderposition * encodercounts
        
    Args:
        port (str): Either a COMport designation ("COM42") or a IP address "192.168.1.42"
        encodercounts (list of int): Conversion factor from unit of interest
            
    Examples:
        my_stage = StageControlGalil('COM19', [2,2,2])
        my_stage2 = StageControlGalil('192.168.1.45', [2,2,2])
    
    Todo
        * correct tracking of the internal and external absolute positions: Do everything via the Galil controller
        * is it truly a good idea to set the stage to zero for each new instantiation?
    
    '''

    def __init__(self,port,encodercounts = [2,2,2]):
        super().__init__()
        
        self.port = port
        self.encodercounts = encodercounts
        '''The number of '''
        self.num_axes = len(encodercounts)
        
        self.g = gclib.py()
        
        '''Opens connection to the stage controller'''
        if port[0:3] == 'COM':
            self.port = port
            self.baudrate = '19200'
            self.timeout = '60000'
            self.connectionstring = self.COMport + ' --baud ' + self.baudrate + ' --subscribe ALL --timeout ' + self.timeout
            self.g.GOpen(self.connectionstring)
        else:
            '''Assumes Ethernet connection'''
            self.connectionstring = self.port + ' --direct'
            self.g.GOpen(self.connectionstring)
        
        '''To address the stages for commands, the list of axes designations needs 
        to be truncated.'''
        _axisstring = 'ABCDEFGH'
        self._axistring = _axisstring[0:self.num_axes]  
        
        '''Create an default position list with the right length'''
        self.positions = [0 for i in range(self.num_axes)]
        self.read_position()
        
        self.initflag = True
        self.unit = 'micron'
        self.speed = 5000
        self.slow_speed = 5000
        self.fast_speed = 20000
        self.set_speed(self.slow_speed)
        '''
        self.read_position('x')
        self.read_position('y')
        self.read_position('z')
        '''
        self.initflag = False

    def close(self):
        '''Closes connection to the stage'''
        self.g.GClose()

    def controller_info(self):
        '''Returns COM port/IP adress, DMC version and serial number of the controller'''
        return self.g.GInfo()
    
    def _send_command(self, command):
        '''Sends a command to the controller
        
        Try-except block included to catch errors - this is dangerous - no checking of success 
        
        Args:
            command (str): Command string to be sent
            
        Returns:
            answer (str): Answer by the controller
        '''
        try:
            return self.g.GCommand(command)
        except Exception as error:
            logger.exception(error)
            
    def stop(self, restart_programs=False):
        '''Stops movement on all axes by sending ST to the controller
        
        After ST, also program execution stops -- using the restart_programs 
        flag, they can be restarted this is risky, here it 
        is assumed that any program runs only the handcontroller and does 
        not induce any movement.
        '''
        self._send_command('ST')
        if restart_programs == True:
            self.execute_program()
    
    def execute_program(self):
        self._send_command('XQ')
            
    def wait_until_done(self, axis):
        '''Requires X,Y,Z or A,B,C etc..'''
        self.g.GMotionComplete(axis.upper())
        
    def read_position(self):
        '''Reports position from the stages 
        
        Notes:
            Updates internal position list self.positions as well.
        
        Returns:
            positions (list): list of positions 
        
        '''
        position = self._send_command('RP')
        position_list = position.split(',')
        try:
            position_list = [int(float(i)) for i in position_list]
            self.positions = [item[0]/item[1] for item in zip(position_list,self.encodercounts)]
        except Exception as error:
            logger.exception(error)

        return self.positions
    
    def move_relative(self, motion_dict):
        '''Command for relative motion 
        
        Args:
            motion_dict (dict): Dictionary in the form {1: 4000, 2:-234}, in general {axis_id:requested_position} 
                                with axis_id (int) and requested_position (int).
        Returns:
            Carries out motion
        '''
        
        ''' Set speed according to requested distance unless it has been set before '''
        if any(abs(distance)>250 for distance in motion_dict.values()):
            if self.speed != self.slow_speed:
                self.set_speed(self.slow_speed)
        else:
            if self.speed != self.fast_speed:
                self.set_speed(self.fast_speed)
                
        ''' Convert um to encodercounts '''
        command_list = self._convert_dict_to_sorted_list(motion_dict, self.num_axes)
        encoder_command_list = [i[0]*i[1] for i in zip(command_list, self.encodercounts)]
                
        ''' Create command string and initiate movement 
        
        Zero padding is important here as the non-commanded axes should stay where they are,
        otherwise previous relative motion commands are retained and possibly excuted 
        again after the 'BG' command ist sent.
        '''                
        command_string = self._convert_sorted_list_to_galil_string(encoder_command_list, zero_padding=True)
        self._send_command('PR'+command_string)
        self._send_command('BG')         
    
    def move_absolute(self, motion_dict):
        ''' Command for absolute motion 
        
        Args:
            motion_dict (dict): Dictionary in the form {1: 4000, 2:-234}, in general {axis_id:requested_position} 
                                with axis_id (int) and requested_position (int).
        
        '''
        
        ''' Set speed according to requested distance unless it has been set before '''
        current_positions = self.read_position()
        requested_positions = self._convert_dict_to_sorted_list(motion_dict, self.num_axes)
        ''' Set requested positions to current value to make the list difference zero
        Otherwise, the distances are not detected correctly
        '''
        requested_positions = [pos[0] if pos[0] != 0 else pos[1] for pos in zip(requested_positions,current_positions)]
        position_differences = self._list_difference(requested_positions,current_positions)    
        if any(abs(distance)>250 for distance in position_differences):
            if self.speed != self.slow_speed:
                self.set_speed(self.slow_speed)
        else:
            if self.speed != self.fast_speed:
                self.set_speed(self.fast_speed)
                
        ''' Convert um to encodercounts '''
        command_list = self._convert_dict_to_sorted_list(motion_dict, self.num_axes)
        encoder_command_list = [i[0]*i[1] for i in zip(command_list, self.encodercounts)]
                
        ''' Carry out motion 
        
        Removing zero padding is important here as the non-commanded axes should stay where they are,
        otherwise they move to position 0.
        '''
        command_string = self._convert_sorted_list_to_galil_string(encoder_command_list, zero_padding=False)
        self._send_command('PA'+command_string)
        self._send_command('BG')         
    
    def set_speed(self, speed):
        '''Sets speed for all axes to the requested value 
        
        Args:
            speed (int): speed value 
        
        '''
        list_axes = [i+1 for i in range(self.num_axes)]
        command_dict = {i : speed for i in list_axes}
        command_string = self._convert_dict_to_galil_string(command_dict, self.num_axes)
        self._send_command('SP'+command_string)
        self.speed = speed
        
    def get_speed(self):
        return self.speed
    
    def _convert_dict_to_galil_string(self, command_dict, length, zero_padding=False):
        '''Converts a dict of the form {1: 4000, 3:-234} into a string '4000,,-234'
        which can be combined with a command ('RP4000,,-234') and sent to the stage controller.

        Args:
            command_dict (dict): Command dictionary in the form {key (int): command value (int)}
            length (int): Length of output items
            zero_padding (bool): Indicates whether the values between commas should be filled with zeros

        Returns:
            galil_string (str): String with values between separating commas
        '''
        keylist = list(command_dict.keys())
        keylist.sort()

        galil_string = ''

        for i in range(length):
            if i+1 in keylist:
                galil_string += str(command_dict[i+1])
                galil_string += ','
            else:
                if zero_padding == True:
                    galil_string += '0'
                galil_string += ','
        return galil_string    
        
    def _convert_sorted_list_to_galil_string(self, command_list, zero_padding=True):
        '''Converts a list of the form [80,123,-53] into a string '80,123,-53'
        which can be combined with a command ('RP80,123,-53') and sent to the stage controller.
        
        Notes:
            If zero_padding is set to False, zeros are removed from the output: [80,0,-53] --> '80,,-53'

        Args:
            command_list (list): Command list in the form [80,123,-53]
            length (int): Length of output items
            zero_padding (bool): Indicates whether the values between commas should be filled with zeros

        Returns:
            galil_string (str): String with values between separating commas
        '''
        galil_string = ''
        for i in range(len(command_list)):
            if command_list[i] == 0 and zero_padding == False:
                galil_string += ','
            else:
                galil_string += str(command_list[i])
                galil_string += ','
            
        return galil_string

    def _convert_dict_to_sorted_list(self, input_dict, length):
        '''Converts an input dictionary into a zero-padded list of a certain length
        
        Args:
            input_dict (dict): Input dictionary in the form {key (int): value (int)} .
            length (int): Number of elements required in the output list.
        Returns:
            List with n = length sorted elements which are sorted
        '''
        keylist = list(input_dict.keys())
        keylist.sort()
        return [input_dict[i+1] if i+1 in keylist else 0 for i in range(length)]
    
    def _list_difference(self, list1, list2):
        '''Returns the element-wise difference of two lists of the same length
        
        Args:
            list1 (list): First list of values
            list2 (list): Second list of values
        '''
        return [i[0]-i[1] for i in zip(list1, list2)]
